# Copyright 2015 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Brocade generator."""

from aerleon.lib import cisco


class Brocade(cisco.Cisco):
    """A brocade policy object.

    Brocade devices do not like protocol numbers. Revert the protocol numbers to
    names just before emitting acl lines to minimize difference from Cisco logic.
    """

    _PLATFORM = 'brocade'
    SUFFIX = '.bacl'
    # Protocols should be emitted as they were in the policy (names).
    _PROTO_INT = False
    _TERM_REMARK = False



# Copyright 2016 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Generic Windows security policy generator; requires subclassing."""

import string
from typing import Union

from absl import logging

from aerleon.lib import aclgenerator, nacaddr
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Header, Policy, Term

CMD_PREFIX = 'netsh ipsec static add '


class Term(aclgenerator.Term):
    """Generate generic windows policy terms."""

    _PLATFORM = 'windows'

    _COMMENT_FORMAT = string.Template(': $comment')

    # filter rules
    _ACTION_TABLE = {}

    def __init__(
        self, term: Term, filter_name: str, filter_action: None, af: str = 'inet'
    ) -> None:
        """Setup a new term.

        Args:
          term: A policy.Term object to represent in windows_ipsec.
          filter_name: The name of the filter chan to attach the term to.
          filter_action: The default action of the filter.
          af: Which address family ('inet' or 'inet6') to apply the term to.

        Raises:
          UnsupportedFilterError: Filter is not supported.
        """
        super().__init__(term)
        self.term = term  # term object
        self.filter = filter_name  # actual name of filter
        self.default_action = filter_action
        self.options = []
        self.af = af

        if af == 'inet6':
            self._all_ips = nacaddr.IPv6('::/0')
        else:
            self._all_ips = nacaddr.IPv4('0.0.0.0/0')

        self.term_name = f'{self.filter[:1]}_{self.term.name}'

    def __str__(self) -> str:
        ret_str = []

        # Don't render icmpv6 protocol terms under inet, or icmp under inet6
        if (self.af == 'inet6' and 'icmp' in self.term.protocol) or (
            self.af == 'inet' and 'icmpv6' in self.term.protocol
        ):
            logging.warning(
                self.NO_AF_LOG_PROTO.substitute(
                    term=self.term.name, proto=self.term.protocol, af=self.af
                )
            )
            return ''

        # append comments to output
        ret_str.append(
            self._COMMENT_FORMAT.substitute(
                filter=self.filter, term=self.term_name, comment=self.term.comment
            )
        )

        # if terms does not specify action, use filter default action
        if not self.term.action:
            self.term.action[0].value = self.default_action

        if self.term.action[0] == 'next':
            return ''

        if len(self.term.action) > 1:
            raise aclgenerator.UnsupportedFilterError(
                '\n%s %s %s %s'
                % (
                    'Multiple actions unsupported by',
                    self._PLATFORM,
                    '\nError in term:',
                    self.term.name,
                )
            )

        # protocol
        if self.term.protocol:
            protocols = self.term.protocol
        else:
            protocols = ['any']

        # addresses
        src_addr = self.term.source_address
        if not src_addr:
            src_addr = [self._all_ips]

        dst_addr = self.term.destination_address
        if not dst_addr:
            dst_addr = [self._all_ips]

        if self.term.source_address_exclude or self.term.destination_address_exclude:
            raise aclgenerator.UnsupportedFilterError(
                '\n%s %s %s %s'
                % (
                    'address exclusions unsupported by',
                    self._PLATFORM,
                    '\nError in term:',
                    self.term.name,
                )
            )

        # ports = Map the ports in a straight list since multiports aren't supported
        (src_ports, dst_ports) = self._HandlePorts(
            self.term.source_port, self.term.destination_port
        )

        # The windows ipsec driver requires either 'tcp' or 'udp' to be specified
        # if a srcport or dstport is specified.  Fail if src or dst ports are
        # specified and of the protocols are not exactly one or both of 'tcp'
        # or 'udp'.
        if (not set(protocols).issubset({'tcp', 'udp'})) and (
            len(src_ports) > 1 or len(dst_ports) > 1
        ):
            raise aclgenerator.UnsupportedFilterError(
                '%s %s %s'
                % (
                    '\n',
                    self.term.name,
                    'src or dst ports may only be specified with "tcp" and/or "udp".',
                )
            )

        # icmp-types
        (icmp_types, protocols) = self._HandleIcmpTypes(self.term.icmp_type, protocols)

        ret_str = []
        self._HandlePreRule(ret_str)
        self._CartesianProduct(
            src_addr, dst_addr, protocols, icmp_types, src_ports, dst_ports, ret_str
        )
        self._HandlePreRule(ret_str)

        return '\n'.join(str(v) for v in ret_str if v)

    def _HandleIcmpTypes(self, icmp_types: list[str], protocols: list[str]) -> tuple[None, None]:
        """Perform implementation-specific icmp_type and protocol transforms.

        Note that icmp_types or protocols are passed as parameters in case they
        are to be munged prior to this function call, and may not be identical
        to self.term.* parameters.

        Args:
          icmp_types:  a list of icmp types, e.g., self.term.icmp_types
          protocols:  a list of protocols, e.g., self.term.protocols

        Returns:
          A pair of lists of (icmp_types, protocols)
        """
        return None, None

    def _HandlePorts(
        self, src_ports: list[tuple[int, int]], dst_ports: list[tuple[int, int]]
    ) -> tuple[None, None]:
        """Perform implementation-specific port transforms.

        Note that icmp_types or protocols are passed as parameters in case they
        are to be munged prior to this function call, and may not be identical
        to self.term.* parameters.

        Args:
          src_ports:  list of source port range tuples, e.g., self.term.source_port
          dst_ports:  list of destination port range tuples

        Returns:
          A pair of lists of (icmp_types, protocols)
        """
        return None, None

    def _HandlePreRule(self, ret_str: list[str]) -> None:
        """Perform any pre-cartesian product transforms on the ret_str array.

        Args:
          ret_str:  an array of strings that will eventually be joined to form
            the string output for the term.
        """
        pass

    def _CartesianProduct(
        self,
        src_addr: list[Union[IPv4, IPv6]],
        dst_addr: list[Union[IPv4, IPv6]],
        protocol: list[str],
        unused_icmp_types: list[str],
        src_port: list[str],
        dst_port: list[str],
        ret_str: list[str],
    ) -> None:
        """Perform any the appropriate cartesian product of the input parameters.

        Args:
          src_addr: a type(IP) list of the source addresses
          dst_addr: a type(IP) list of the destination addresses
          protocol: a string list of the protocols
          icmp_types: a numeric list of the icmp_types
          src_ports: a (start, end) list of the source ports
          dst_ports: a (start,end) list of the destination ports
          ret_str:  an array of strings that will eventually be joined to form
            the string output for the term.
        """
        pass

    def _HandlePostRule(self, ret_str: str) -> None:
        """Perform any port-cartesian product transforms on the ret_str array.

        Args:
          ret_str:  an array of strings that will eventually be joined to form
            the string output for the term.
        """
        pass


class WindowsGenerator(aclgenerator.ACLGenerator):
    """Generates filters and terms from provided policy object."""

    _PLATFORM = 'windows'
    _DEFAULT_PROTOCOL = 'all'
    SUFFIX = '.bat'
    _RENDER_PREFIX = None
    _DEFAULT_ACTION = 'block'
    _TERM = Term

    _GOOD_AFS = ['inet', 'inet6']

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {'option'}
        supported_tokens -= {'verbatim'}

        supported_sub_tokens.update({'action': {'accept', 'deny'}})
        del supported_sub_tokens['option']
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        """Translate a policy from objects into strings."""
        self.windows_policies = []
        default_action = None
        good_default_actions = ['permit', 'block']
        good_options = []

        for header, terms in pol.filters:
            filter_type = None
            filter_options = header.FilterOptions(self._PLATFORM)[1:]
            filter_name = header.FilterName(self._PLATFORM)

            # ensure all options after the filter name are expected
            for opt in filter_options:
                if opt not in good_default_actions + self._GOOD_AFS + good_options:
                    raise aclgenerator.UnsupportedTargetOptionError(
                        '%s %s %s %s'
                        % (
                            '\nUnsupported option found in',
                            self._PLATFORM,
                            'target definition:',
                            opt,
                        )
                    )

            # Check for matching af
            for address_family in self._GOOD_AFS:
                if address_family in filter_options:
                    # should not specify more than one AF in options
                    if filter_type is not None:
                        raise aclgenerator.UnsupportedFilterError(
                            '%s %s %s %s'
                            % (
                                '\nMay only specify one of',
                                self._GOOD_AFS,
                                'in filter options:',
                                filter_options,
                            )
                        )
                    filter_type = address_family
            if filter_type is None:
                filter_type = 'inet'

            # does this policy override the default filter actions?
            for next_target in header.target:
                if next_target.platform == self._PLATFORM:
                    if len(next_target.options) > 1:
                        for arg in next_target.options:
                            if arg in good_default_actions:
                                default_action = arg
            if default_action and default_action not in good_default_actions:
                raise aclgenerator.UnsupportedTargetOptionError(
                    '%s %s %s %s %s'
                    % (
                        '\nOnly',
                        ', '.join(good_default_actions),
                        'default filter action allowed;',
                        default_action,
                        'used.',
                    )
                )

            # add the terms
            new_terms = []
            term_names = set()
            for term in terms:
                if term.name in term_names:
                    raise aclgenerator.DuplicateTermError(
                        f'You have a duplicate term: {term.name}'
                    )
                term_names.add(term.name)

                if 'established' in term.option or 'tcp-established' in term.option:
                    continue
                new_terms.append(self._TERM(term, filter_name, default_action, filter_type))

            self.windows_policies.append(
                (header, filter_name, filter_type, default_action, new_terms)
            )

    def __str__(self) -> str:
        target = []
        pretty_platform = f'{self._PLATFORM[0].upper()}{self._PLATFORM[1:]}'

        if self._RENDER_PREFIX:
            target.append(self._RENDER_PREFIX)

        for header, _, filter_type, default_action, terms in self.windows_policies:
            # Add comments for this filter
            target.append(f': {pretty_platform} {header.FilterName(self._PLATFORM)} Policy')

            self._HandlePolicyHeader(header, target)

            # reformat long text comments, if needed
            comments = aclgenerator.WrapWords(header.comment, 70)
            if comments and comments[0]:
                for line in comments:
                    target.append(f': {line}')
                target.append(':')
            # add the p4 tags
            target.extend(aclgenerator.AddRepositoryTags(': '))
            target.append(f": {filter_type}")

            if default_action:
                raise aclgenerator.UnsupportedTargetOptionError(
                    'Windows generator does not support default actions'
                )

            # add the terms
            for term in terms:
                term_str = str(term)
                if term_str:
                    target.append(term_str)
                    self._HandleTermFooter(header, term, target)

        target.append('')
        return '\n'.join(target)

    def _HandlePolicyHeader(self, header: Header, target: list[str]) -> None:
        pass

    def _HandleTermFooter(self, header: Header, term: Term, target: list[str]) -> None:
        pass



# Copyright 2015 The Capirca Project Authors All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Nsxv generator."""

import re
import xml
from typing import Optional

from absl import logging

from aerleon.lib import aclgenerator, nacaddr, policy

_ACTION_TABLE = {
    'accept': 'allow',
    'deny': 'deny',
    'reject': 'reject',
    'reject-with-tcp-rst': 'reject',  # tcp rst not supported
}

_XML_TABLE = {
    'actionStart': '<action>',
    'actionEnd': '</action>',
    'srcIpv4Start': '<source><type>Ipv4Address</type><value>',
    'srcIpv4End': '</value></source>',
    'destIpv4Start': '<destination><type>Ipv4Address</type><value>',
    'destIpv4End': '</value></destination>',
    'protocolStart': '<protocol>',
    'protocolEnd': '</protocol>',
    'serviceStart': '<service>',
    'serviceEnd': '</service>',
    'appliedToStart': '<appliedTo><type>SecurityGroup</type><value>',
    'appliedToEnd': '</value></appliedTo>',
    'srcPortStart': '<sourcePort>',
    'srcPortEnd': '</sourcePort>',
    'destPortStart': '<destinationPort>',
    'destPortEnd': '</destinationPort>',
    'icmpTypeStart': '<subProtocol>',
    'icmpTypeEnd': '</subProtocol>',
    'logTrue': '<loggingEnabled>true</loggingEnabled>',
    'logFalse': '<loggingEnabled>false</loggingEnabled>',
    'sectionStart': '<section>',
    'sectionEnd': '</section>',
    'nameStart': '<name>',
    'nameEnd': '</name>',
    'srcIpv6Start': '<source><type>Ipv6Address</type><value>',
    'srcIpv6End': '</value></source>',
    'destIpv6Start': '<destination><type>Ipv6Address</type><value>',
    'destIpv6End': '</value></destination>',
    'noteStart': '<notes>',
    'noteEnd': '</notes>',
}

_NSXV_SUPPORTED_KEYWORDS = [
    'name',
    'action',
    'comment',
    'destination_address',
    'destination_address_exclude',
    'destination_port',
    'expiration',
    'icmp_type',
    'protocol',
    'source_address',
    'source_address_exclude',
    'source_port',
    'expiration',
    'logging',
]


# generic error class
class Error(aclgenerator.Error):
    """Generic error class."""

    pass


class UnsupportedNsxvAccessListError(Error):
    """Raised when we're give a non named access list."""

    pass


class NsxvAclTermError(Error):
    """Raised when there is a problem in a nsxv access list."""

    pass


class NsxvDuplicateTermError(Error):
    """Raised when there is a duplicate."""

    pass


class Term(aclgenerator.Term):
    """Creates a  single ACL Term for Nsxv."""

    def __init__(
        self, term: policy.Term, filter_type: str, applied_to: Optional[str] = None, af: int = 4
    ):
        self.term = term
        # Our caller should have already verified the address family.
        assert af in (4, 6)
        self.af = af
        self.filter_type = filter_type
        self.applied_to = applied_to

    def __str__(self) -> str:
        """Convert term to a rule string.

        Returns:
          A rule as a string.

        Raises:
          NsxvAclTermError: When unknown icmp-types are specified

        """
        ret_str = ['']

        # Don't render icmpv6 protocol terms under inet, or icmp under inet6
        if (self.af == 6 and 'icmp' in self.term.protocol) or (
            self.af == 4 and 'icmpv6' in self.term.protocol
        ):
            if self.filter_type != 'mixed':
                logging.warning(
                    self.NO_AF_LOG_PROTO.substitute(
                        term=self.term.name,
                        proto=', '.join(self.term.protocol),
                        af=self.filter_type,
                    )
                )
            return ''

        # Term verbatim is not supported
        if self.term.verbatim:
            raise NsxvAclTermError('Verbatim are not implemented in standard ACLs')

        # Term option is not supported
        if self.term.option:
            for opt in [str(single_option) for single_option in self.term.option]:
                if (opt.find('tcp-established') == 0) or (opt.find('established') == 0):
                    return ''
                else:
                    raise NsxvAclTermError('Option are not implemented in standard ACLs')

        # check for keywords Nsxv does not support
        term_keywords = self.term.__dict__
        unsupported_keywords = []
        for key in term_keywords:
            if term_keywords[key]:
                # translated is obj attribute not keyword
                if ('translated' not in key) and (key not in _NSXV_SUPPORTED_KEYWORDS):
                    unsupported_keywords.append(key)
        if unsupported_keywords:
            logging.warning(
                'WARNING: The keywords %s in Term %s are not supported ' 'in Nsxv ',
                unsupported_keywords,
                self.term.name,
            )

        name = '{}{}{}'.format(
            _XML_TABLE.get('nameStart'), self.term.name, _XML_TABLE.get('nameEnd')
        )

        notes = ''
        if self.term.comment:
            for comment in self.term.comment:
                notes = f'{notes}{comment}'
            notes = f"{_XML_TABLE.get('noteStart')}{notes}{_XML_TABLE.get('noteEnd')}"

        # protocol
        protocol = None

        if self.term.protocol:
            protocol = list(map(self.PROTO_MAP.get, self.term.protocol, self.term.protocol))

            # icmp-types
            icmp_types = ['']
            if self.term.icmp_type:
                icmp_types = self.NormalizeIcmpTypes(
                    self.term.icmp_type, self.term.protocol, self.af
                )

        # for mixed filter type get both IPV4address and IPv6Address
        af_list = []
        if self.filter_type == 'mixed':
            af_list = [4, 6]
        else:
            af_list = [self.af]

        source_address = None
        destination_address = None
        source_addr = []
        destination_addr = []

        source_v4_addr = []
        source_v6_addr = []
        dest_v4_addr = []
        dest_v6_addr = []

        for af in af_list:
            # source address
            if self.term.source_address:
                source_address = self.term.GetAddressOfVersion('source_address', af)
                source_address_exclude = self.term.GetAddressOfVersion(
                    'source_address_exclude', af
                )
                if source_address_exclude:
                    source_address = nacaddr.ExcludeAddrs(source_address, source_address_exclude)

                if source_address:
                    if af == 4:
                        source_v4_addr = source_address
                    else:
                        source_v6_addr = source_address
                source_addr = source_v4_addr + source_v6_addr

            # destination address
            if self.term.destination_address:
                destination_address = self.term.GetAddressOfVersion('destination_address', af)
                destination_address_exclude = self.term.GetAddressOfVersion(
                    'destination_address_exclude', af
                )
                if destination_address_exclude:
                    destination_address = nacaddr.ExcludeAddrs(
                        destination_address, destination_address_exclude
                    )

                if destination_address:
                    if af == 4:
                        dest_v4_addr = destination_address
                    else:
                        dest_v6_addr = destination_address
                destination_addr = dest_v4_addr + dest_v6_addr

        # Check for mismatch IP for source and destination address for mixed filter
        if self.filter_type == 'mixed':
            if source_addr and destination_addr:
                if source_v4_addr and not dest_v4_addr:
                    source_addr = source_v6_addr
                elif source_v6_addr and not dest_v6_addr:
                    source_addr = source_v4_addr
                elif dest_v4_addr and not source_v4_addr:
                    destination_addr = dest_v6_addr
                elif dest_v6_addr and not source_v6_addr:
                    destination_addr = dest_v4_addr

                if not source_addr or not destination_addr:
                    logging.warning(
                        'Term %s will not be rendered as it has IPv4/IPv6 '
                        'mismatch for source/destination for mixed address '
                        'family.',
                        self.term.name,
                    )
                    return ''

        # ports
        source_port = None
        destination_port = None
        if self.term.source_port:
            source_port = self.term.source_port
        if self.term.destination_port:
            destination_port = self.term.destination_port

        # logging
        log = 'false'
        if self.term.logging:
            log = 'true'

        sources = ''
        if source_addr:
            sources = '<sources excluded="false">'
            for saddr in source_addr:
                # inet4
                if isinstance(saddr, nacaddr.IPv4):
                    if saddr.num_addresses > 1:
                        saddr = '{}{}{}'.format(
                            _XML_TABLE.get('srcIpv4Start'),
                            saddr.with_prefixlen,
                            _XML_TABLE.get('srcIpv4End'),
                        )
                    else:
                        saddr = '{}{}{}'.format(
                            _XML_TABLE.get('srcIpv4Start'),
                            saddr.network_address,
                            _XML_TABLE.get('srcIpv4End'),
                        )
                    sources = f'{sources}{saddr}'
                # inet6
                if isinstance(saddr, nacaddr.IPv6):
                    if saddr.num_addresses > 1:
                        saddr = '{}{}{}'.format(
                            _XML_TABLE.get('srcIpv6Start'),
                            saddr.with_prefixlen,
                            _XML_TABLE.get('srcIpv6End'),
                        )
                    else:
                        saddr = '{}{}{}'.format(
                            _XML_TABLE.get('srcIpv6Start'),
                            saddr.network_address,
                            _XML_TABLE.get('srcIpv6End'),
                        )
                    sources = f'{sources}{saddr}'
            sources = f'{sources}</sources>'

        destinations = ''
        if destination_addr:
            destinations = '<destinations excluded="false">'
            for daddr in destination_addr:
                # inet4
                if isinstance(daddr, nacaddr.IPv4):
                    if daddr.num_addresses > 1:
                        daddr = '{}{}{}'.format(
                            _XML_TABLE.get('destIpv4Start'),
                            daddr.with_prefixlen,
                            _XML_TABLE.get('destIpv4End'),
                        )
                    else:
                        daddr = '{}{}{}'.format(
                            _XML_TABLE.get('destIpv4Start'),
                            daddr.network_address,
                            _XML_TABLE.get('destIpv4End'),
                        )
                    destinations = f'{destinations}{daddr}'
                # inet6
                if isinstance(daddr, nacaddr.IPv6):
                    if daddr.num_addresses > 1:
                        daddr = '{}{}{}'.format(
                            _XML_TABLE.get('destIpv6Start'),
                            daddr.with_prefixlen,
                            _XML_TABLE.get('destIpv6End'),
                        )
                    else:
                        daddr = '{}{}{}'.format(
                            _XML_TABLE.get('destIpv6Start'),
                            daddr.network_address,
                            _XML_TABLE.get('destIpv6End'),
                        )
                    destinations = f'{destinations}{daddr}'
            destinations = f'{destinations}</destinations>'

        services = []
        if protocol:
            services.append('<services>')
            for proto in protocol:
                if proto != 'any':
                    services.append(
                        self._ServiceToString(proto, source_port, destination_port, icmp_types)
                    )
            services.append('</services>')

        service = ''
        for s in services:
            service = f'{service}{s}'

        # applied_to
        applied_to_list = ''
        if self.applied_to:
            applied_to_list = '<appliedToList>'
            applied_to_element = '{}{}{}'.format(
                _XML_TABLE.get('appliedToStart'),
                self.applied_to,
                _XML_TABLE.get('appliedToEnd'),
            )
            applied_to_list = f'{applied_to_list}{applied_to_element}'
            applied_to_list = f'{applied_to_list}</appliedToList>'

        # action
        action = '{}{}{}'.format(
            _XML_TABLE.get('actionStart'),
            _ACTION_TABLE.get(str(self.term.action[0])),
            _XML_TABLE.get('actionEnd'),
        )

        ret_lines = []
        ret_lines.append(
            '<rule logged="%s">%s%s%s%s%s%s%s</rule>'
            % (log, name, action, sources, destinations, service, applied_to_list, notes)
        )

        # remove any trailing spaces and replace multiple spaces with singles
        stripped_ret_lines = [re.sub(r'\s+', ' ', x).rstrip() for x in ret_lines]
        ret_str.extend(stripped_ret_lines)
        return ''.join(ret_str)

    def _ServiceToString(
        self, proto: int, sports: tuple[int, int], dports: tuple[int, int], icmp_types: list[str]
    ):
        """Converts service to string.

        Args:
          proto: str, protocl
          sports: str list or none, the source port
          dports: str list or none, the destination port
          icmp_types: icmp-type numeric specification (if any)

        Returns:
          Service definition.

        """
        service = ''
        # for icmp and icmpv6
        if proto == 1 or proto == 58:
            # handle icmp protocol
            for icmp_type in icmp_types:
                icmp_service = '{}{}{}{}'.format(
                    _XML_TABLE.get('serviceStart'),
                    _XML_TABLE.get('protocolStart'),
                    proto,
                    _XML_TABLE.get('protocolEnd'),
                )
                # handle icmp types
                if icmp_type:
                    icmp_type = '{}{}{}'.format(
                        _XML_TABLE.get('icmpTypeStart'),
                        str(icmp_type),
                        _XML_TABLE.get('icmpTypeEnd'),
                    )
                    icmp_service = f'{icmp_service}{icmp_type}'
                icmp_service = f"{icmp_service}{_XML_TABLE.get('serviceEnd')}"
                service = f'{service}{icmp_service}'
        else:
            # handle other protocols
            service = '{}{}{}{}'.format(
                _XML_TABLE.get('serviceStart'),
                _XML_TABLE.get('protocolStart'),
                proto,
                _XML_TABLE.get('protocolEnd'),
            )

            # handle source ports
            if sports:
                str_sport = []
                for sport in sports:
                    if sport[0] != sport[1]:
                        str_sport.append(f'{sport[0]}-{sport[1]}')
                    else:
                        str_sport.append(f'{sport[0]}')
                service = '{}{}{}{}'.format(
                    service,
                    _XML_TABLE.get('srcPortStart'),
                    ', '.join(str_sport),
                    _XML_TABLE.get('srcPortEnd'),
                )

            # handle destination ports
            if dports:
                str_dport = []
                for dport in dports:
                    if dport[0] != dport[1]:
                        str_dport.append(f'{dport[0]}-{dport[1]}')
                    else:
                        str_dport.append(f'{dport[0]}')
                service = '{}{}{}{}'.format(
                    service,
                    _XML_TABLE.get('destPortStart'),
                    ', '.join(str_dport),
                    _XML_TABLE.get('destPortEnd'),
                )
            service = f"{service}{_XML_TABLE.get('serviceEnd')}"

        return service


class Nsxv(aclgenerator.ACLGenerator):
    """Nsxv rendering class.

      This class takes a policy object and renders the output into a syntax
      which is understood by nsxv policy.

    Attributes:
      pol: policy.Policy object

    Raises:
    UnsupportedNsxvAccessListError: Raised when we're give a non named access
    list.

    """

    _PLATFORM = 'nsxv'
    _DEFAULT_PROTOCOL = 'ip'
    SUFFIX = '.nsx'

    _OPTIONAL_SUPPORTED_KEYWORDS = {
        'expiration',
        'logging',
    }
    _FILTER_OPTIONS_DICT = {}

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {'logging'}
        supported_sub_tokens.update(
            {'action': {'accept', 'deny', 'reject', 'reject-with-tcp-rst'}}
        )
        del supported_sub_tokens['option']
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int):
        self.nsxv_policies = []
        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            if len(filter_options) >= 2:
                filter_name = filter_options[1]

            # get filter type, section id and applied To
            self._ParseFilterOptions(filter_options)

            filter_type = self._FILTER_OPTIONS_DICT['filter_type']
            applied_to = self._FILTER_OPTIONS_DICT['applied_to']

            term_names = set()
            new_terms = []
            for term in terms:
                # Check for duplicate terms
                if term.name in term_names:
                    raise NsxvDuplicateTermError(f'There are multiple terms named: {term.name}')
                term_names.add(term.name)

                # Get the mapped action value
                # If there is no mapped action value term is not rendered
                mapped_action = _ACTION_TABLE.get(str(term.action[0]))
                if not mapped_action:
                    logging.warning(
                        'WARNING: Action %s in Term %s is not valid and ' 'will not be rendered.',
                        term.action,
                        term.name,
                    )
                    continue

                term.name = self.FixTermLength(term.name)

                if filter_type == 'inet':
                    af = 'inet'
                    term = self.FixHighPorts(term, af=af)
                    if not term:
                        continue
                    new_terms.append(Term(term, filter_type, applied_to, 4))

                if filter_type == 'inet6':
                    af = 'inet6'
                    term = self.FixHighPorts(term, af=af)
                    if not term:
                        continue
                    new_terms.append(Term(term, filter_type, applied_to, 6))

                if filter_type == 'mixed':
                    if 'icmpv6' not in term.protocol:
                        inet_term = self.FixHighPorts(term, 'inet')
                        if not inet_term:
                            continue
                        new_terms.append(Term(inet_term, filter_type, applied_to, 4))
                    else:
                        inet6_term = self.FixHighPorts(term, 'inet6')
                        if not inet6_term:
                            continue
                        new_terms.append(Term(inet6_term, filter_type, applied_to, 6))

            self.nsxv_policies.append((header, filter_name, [filter_type], new_terms))

    def _ParseFilterOptions(self, filter_options: list[str]):
        """Parses the target in header for filter type, section_id and applied_to.

        Args:
          filter_options: list of remaining target options

        Returns:
          A dictionary that contains fields necessary to create the firewall
          rule.

        Raises:
          UnsupportedNsxvAccessListError: Raised when we're give a non named access
          list.
        """
        # check for filter type
        if not 2 <= len(filter_options) <= 5:
            raise UnsupportedNsxvAccessListError(
                'Invalid Number of options specified: %d. Required options '
                'are: filter type and section name. Platform: %s'
                % (len(filter_options), self._PLATFORM)
            )
        # mandatory section_name
        section_name = filter_options[0]
        # mandatory
        filter_type = filter_options[1]

        # a mixed filter outputs both ipv4 and ipv6 acls in the same output file
        good_filters = ['inet', 'inet6', 'mixed']

        # check if filter type is renderable
        if filter_type not in good_filters:
            raise UnsupportedNsxvAccessListError(
                'Access list type %s not supported by %s (good types: %s)'
                % (filter_type, self._PLATFORM, str(good_filters))
            )

        section_id = 0
        applied_to = None
        filter_opt_len = len(filter_options)

        if filter_opt_len > 2:
            for index in range(2, filter_opt_len):
                if index == 2 and filter_options[2] != 'securitygroup':
                    section_id = filter_options[2]
                    continue
                if filter_options[index] == 'securitygroup':
                    if index + 1 <= filter_opt_len - 1:
                        applied_to = filter_options[index + 1]
                        break
                    else:
                        raise UnsupportedNsxvAccessListError(
                            f'Security Group Id is not provided for {self._PLATFORM}'
                        )

        self._FILTER_OPTIONS_DICT['section_name'] = section_name
        self._FILTER_OPTIONS_DICT['filter_type'] = filter_type
        self._FILTER_OPTIONS_DICT['section_id'] = section_id
        self._FILTER_OPTIONS_DICT['applied_to'] = applied_to

    def __str__(self) -> str:
        """Render the output of the Nsxv policy."""

        target_header = []
        target = []

        # add the p4 tags
        target.append('<!--')
        target.extend(aclgenerator.AddRepositoryTags('\n'))
        target.append('\n')
        target.append('-->')

        for _, _, _, terms in self.nsxv_policies:
            section_name = self._FILTER_OPTIONS_DICT['section_name']
            # check section id value
            section_id = self._FILTER_OPTIONS_DICT['section_id']
            if not section_id or section_id == 0:
                logging.warning(
                    'WARNING: Section-id is 0. A new Section is created '
                    'for %s. If there is any existing section, it '
                    'will remain unreferenced and should be removed '
                    'manually.',
                    section_name,
                )
                target.append('<section name="%s">' % (section_name.strip(' \t\n\r')))
            else:
                target.append(
                    '<section id="{}" name="{}">'.format(section_id, section_name.strip(' \t\n\r'))
                )

            # now add the terms
            for term in terms:
                term_str = str(term)
                if term_str:
                    target.append(term_str)

            # ensure that the header is always first
            target = target_header + target
            target.append(f"{_XML_TABLE.get('sectionEnd')}")
            target.append('\n')

            target_as_xml = xml.dom.minidom.parseString(''.join(target))
        # TODO(robankeny) utf encoding with refactoring after migration to py3
        return target_as_xml.toprettyxml(indent='  ')



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Simple util to grep through network and service definitions.

Examples:
  To find out which tokens contain "10.4.3.1" use
  $ cgrep.py -i 10.4.3.1

  To find out if token 'FOO' includes ip "1.2.3.4" use
  $ cgrep.py -t FOO -i 1.2.3.4

  To find the difference and union of tokens 'FOO' and 'BAR' use
  $ cgrep.py -c FOO BAR

  To find the difference of network tokens to which 2 IPs belong use
  $ cgrep.py -g 1.1.1.1 2.2.2.2

  To find which IPs are in the 'FOO' network token use
  $ cgrep.py -o FOO

  To find which port & protocol pairs are in a service token 'FOO' use
  $ cgrep.py -s FOO

  To find which service tokens contain port '22' and protocol 'tcp' use
  $ cgrep.py -p 22 tcp
"""


import argparse
import logging
import pprint

from aerleon.lib import nacaddr, naming


def is_valid_ip(arg):
    """Validates a value to be an IP or not.

    Args:
      arg: potential IP address as a string.

    Returns:
      arg as IP object (if arg is an IP)

    Raises:
      Error (if arg is not an IP)
    """
    try:
        nacaddr.IP(arg)
    except:
        raise argparse.ArgumentTypeError(f'{arg} is an invalid ip address')
    return arg


def cli_options():
    """Builds the argparse options for cgrep.

    TODO(robankeny): Move this to flags.

    Returns:
      parser: the arguments, ready to be parsed.
    """

    parser = argparse.ArgumentParser(
        description='c[apirca]grep', formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        '-d',
        '--def',
        '--definitions_directory',
        dest='defs',
        help='Network Definitions directory location. \n',
        default='./def',
    )

    # -i and -t can be used together, but not with any other option.
    ip_group = parser.add_argument_group()
    # take 1 or more IPs
    ip_group.add_argument(
        '-i',
        '--ip',
        dest='ip',
        nargs='+',
        type=is_valid_ip,
        help='Return list of definitions containing the ' 'IP(s).\nMultiple IPs permitted.',
    )

    ip_group.add_argument(
        '-t',
        '--token',
        dest='token',
        help=(
            'See if an IP is contained within the given '
            'token.\nMust be used in conjunction with '
            '-i/--ip [addr].'
        ),
    )

    exclusive_group = parser.add_mutually_exclusive_group()
    # the rest of the arguments are mutually exclusive with each other,
    # and -i / -t
    exclusive_group.add_argument(
        '-c',
        '--cmp',
        dest='cmp',
        nargs=2,
        metavar=('OBJ', 'OBJ'),
        help=('Compare the two given network ' 'definition tokens'),
    )

    exclusive_group.add_argument(
        '-g',
        '--gmp',
        dest='gmp',
        nargs=2,
        type=is_valid_ip,
        metavar=('IP', 'IP'),
        help=('Diff the network objects to' ' which the given IP(s) belong'),
    )

    exclusive_group.add_argument(
        '-o',
        '--obj',
        dest='obj',
        nargs='+',
        help=('Return list of IP(s) contained within ' 'the given token(s)'),
    )

    exclusive_group.add_argument(
        '-s',
        '--svc',
        dest='svc',
        nargs='+',
        help=('Return list of port(s) contained ' 'within given token(s)'),
    )

    exclusive_group.add_argument(
        '-p',
        '--port',
        dest='port',
        nargs=2,
        metavar=('PORT', 'PROTO'),
        help=('Returns a list of tokens containing ' 'the given port and protocol'),
    )

    return parser


def main(parser):
    """Determines the code path based on the arguments passed.

    Args:
      parser: the argument parser, but not parsed yet.
    """
    logging.basicConfig(level=logging.INFO)
    options = parser.parse_args()
    db = naming.Naming(options.defs)
    p = pprint.PrettyPrinter(indent=1, depth=4, width=1).pprint

    # if -i and any other option:
    if options.ip and any([options.gmp, options.cmp, options.obj, options.svc, options.port]):
        logging.info('You can only use -i with -t or by itself')

    # if -i and -t
    elif options.token and options.ip:
        try:
            get_nets([options.token], db)
        except naming.UndefinedAddressError:
            logging.info("Network group '%s' is not defined!", options.token)
        else:
            results = compare_ip_token(options, db)
            logging.info(results)

    # if -t, but not -i; invalid!
    elif options.token and not options.ip:
        logging.info('You must specify an IP Address with -i [addr]')

    # if -i
    elif options.ip:
        for ip in options.ip:
            groups = get_ip_parents(ip, db)
            logging.info('Results for IP: %s', ip)
            # iterate and print the tokens we found.
            for name, networks in groups:
                # print the group name [0], and the networks it was in [1]
                logging.info('%s  %s', name, networks)

    elif options.gmp:
        common, diff1, diff2 = group_diff(options, db)
        print_diff(options.gmp[0], common, diff1, diff2)
        logging.info('')
        print_diff(options.gmp[1], common, diff2, diff1)

    # if -c
    elif options.cmp:
        meta, results = compare_tokens(options, db)
        first_name = meta[0]
        second_name = meta[1]
        union = meta[2]
        logging.info('Union of %s and %s:\n %s\n', first_name, second_name, union)
        logging.info('Diff of %s and %s:', first_name, second_name)
        for i in results:
            logging.info(f" {i}")
        logging.info('')
        first_obj, sec_obj = options.cmp
        if check_encapsulated('network', first_obj, sec_obj, db):
            logging.info('%s fully encapsulates %s', sec_obj, first_obj)
        else:
            logging.info('%s does _not_ fully encapsulate %s', sec_obj, first_obj)
        # check the other way around.
        if check_encapsulated('network', sec_obj, first_obj, db):
            logging.info('%s fully encapsulates %s', first_obj, sec_obj)
        else:
            logging.info('%s does _not_ fully encapsulate %s', first_obj, sec_obj)

    # if -o
    elif options.obj:
        for obj in options.obj:
            try:
                token, ips = get_nets([obj], db)[0]
            except naming.UndefinedAddressError:
                logging.info('%s is an invalid object', obj)
            else:
                logging.info(f"{token}:")
                # convert list of ip objects to strings and sort them
                sorted_ips = nacaddr.SortAddrList(ips)
                p([str(x) for x in sorted_ips])

    # if -s
    elif options.svc:
        try:
            results = get_ports(options.svc, db)
        except naming.UndefinedServiceError:
            logging.info('%s contains an invalid service object', str(options.svc))
        else:
            for result in get_ports(options.svc, db):
                svc, port = result
                logging.info(f"{svc}:")
                p(port)

    # if -p
    elif options.port:
        port, protocol, result = get_services(options, db)
        logging.info('%s/%s:', port, protocol)
        p(result)

    # if nothing is passed
    elif not any((options.cmp, options.ip, options.token, options.obj, options.svc, options.port)):
        parser.print_help()
    logging.info('')


def check_encapsulated(obj_type, first_obj, second_obj, db):
    """Checks if a network/service object is entirely contained within another.

    Args:
      obj_type: "network" or "service"
      first_obj: The name of the first network/service object
      second_obj: The name of the secondnetwork/service object
      db: The network and service definitions

    Returns:
      Error or bool:
        ValueError if an invalid object type is passed
        True if the first_obj is entirely within second_obj, otherwise False

    Raises:
      ValueError: When value is not a network or service.
    """
    if obj_type == 'network':
        # the indexing is to get the list of networks out of the tuple[1] and
        # list[0] returned by get_nets
        first = get_nets([first_obj], db)[0][1]
        second = get_nets([second_obj], db)[0][1]

    elif obj_type == 'service':
        first = get_ports([first_obj], db)[0][1]
        second = get_ports([second_obj], db)[0][1]
    else:
        raise ValueError(
            "check_encapsulated() currently only supports "
            "'network' and 'service' for the obj_type parameter"
        )
    # iterates over each object in the first group, and then each obj in the
    # second group, making sure each one in the first is contained
    # somewhere in the second.
    for obj in first:
        for sec_obj in second:
            if obj in sec_obj:
                break
        # if we got through every object in the second group, and didn't have
        # a match, then the first group is not entirely contained.
        else:
            return False
    # if we got here, then the group was fully contained.
    return True


def print_diff(ip, common, diff1, diff2):
    """Print out the common, added, and removed network objects between 2 IPs.

    Args:
      ip: the IP being compared against
      common: the network objects shared between the two IPs
                      ('ip' and the other passed into options.cmp)
      diff1: the network objects present in 'ip' but not in the other IP
                     passed into options.cmp
      diff2: the network objects not present in 'ip' but are present in
                     the other IP passed into options.cmp
    """
    logging.info('IP: %s', ip)
    if common:
        common = [f'  {elem}' for elem in common]
        logging.info('\n'.join(common))
    if diff1:
        diff = [f'+ {elem}' for elem in diff1]
        logging.info('\n'.join(diff))
    if diff2:
        diff = [f'- {elem}' for elem in diff2]
        logging.info('\n'.join(diff))


def group_diff(options, db):
    """Diffs two different group objects.

    Args:
      options: the options sent to the script
      db : network and service definitions

    Returns:
      tuple: the common lines, the differences from 1 to 2,
                            and the differences from 2 to 1
    """
    nested_rvals = []
    for ip in options.gmp:
        nested_rvals.append(get_ip_parents(ip, db))
    # get just the list of groups, stripping out the networks.
    group1 = [x[0] for x in nested_rvals[0]]
    group2 = [x[0] for x in nested_rvals[1]]
    common = list(set(group1) & set(group2))
    diff1 = list(set(group1) - set(group2))
    diff2 = list(set(group2) - set(group1))
    return common, diff1, diff2


def get_ip_parents(ip, db):
    """Gets a list of all network objects that include an IP.

    Args:
      ip: the IP we're looking for the parents of
      db: network and service definitions

    Returns:
      results: a list of all groups that include the IP, in the format:
                       [("Group", ["networks", "matched"]), (etc)]
    """
    results = []
    rval = db.GetIpParents(ip)
    for v in rval:
        nested = db.GetNetParents(v)
        prefix_and_nets = get_nets_and_highest_prefix(ip, v, db)
        if nested:
            for n in nested:
                results.append((f'{n} -> {v}', prefix_and_nets))
        else:
            results.append((v, prefix_and_nets))
    # sort the results by prefix length descending
    results = sorted(results, key=lambda x: x[1][0], reverse=True)
    # strip out the no longer needed prefix lengths before handing off
    for index, group in enumerate(results):
        results[index] = (group[0], group[1][1])
    return results


def get_nets_and_highest_prefix(ip, net_group, db):
    """Find the highest prefix length in all networks given it contains the IP.

    Args:
      ip: the IP address contained in net_group
      net_group: the name of the network object we'll be looking through
      db: network and service definitions

    Returns:
      highest_prefix_length, networks as tuple
        highest_prefix_length : the longest prefix length found,
        networks : network objects
    """
    highest_prefix_length = 0
    networks = []
    ip = nacaddr.IP(ip)
    # loop through all the networks in the net_group
    for net in get_nets([net_group], db)[0][1]:
        # find the highest prefix length for the networks that contain the IP
        if ip.subnet_of(net):
            networks.append(str(net))
            if net.prefixlen > highest_prefix_length:
                highest_prefix_length = net.prefixlen
    return highest_prefix_length, networks


def get_nets(objects, db):
    """Gets a list of all networks that are inside of a network object.

    Args:
      objects: network objects
      db: network and service definitions

    Returns:
      results : all networks inside a network object
    """
    results = []
    for obj in objects:
        net = db.GetNet(obj)
        results.append((obj, net))
    return results


def compare_tokens(options, db):
    """Compares to network objects against each other.

    Args:
      options: the options sent to the script
      db: network and service definitions

    Returns:
      meta, results :
        ((first object, second object, union of those two),
         diff of those two network objects)
    """
    t1, t2 = options.cmp
    d1 = db.GetNet(t1)
    d2 = db.GetNet(t2)
    union = list(set(d1 + d2))
    meta = (t1, t2, union)
    results = []
    for el in set(d1 + d2):
        el = nacaddr.IP(el)
        if el.subnet_of(d1) and el.subnet_of(d2):
            results.append(str(el))
        elif el.subnet_of(d1):
            results.append(str(el))
        elif el.subnet_of(d2):
            results.append(str(el))
    return meta, results


def compare_ip_token(options, db):
    """Looks to see if a network IP is contained in a network object.

    Args:
      options: the options sent to the script
      db: network and service definitions

    Returns:
      results : end-user string stating the results
    """
    token = options.token
    results = []
    for ip in options.ip:
        rval = db.GetIpParents(ip)
        if token in rval:
            results = f'{ip} is in {token}'
        else:
            results = f'{ip} is _not_ in {token}'
    return results


def get_ports(svc_group, db):
    """Gets the ports and protocols defined in a service group.

    Args:
      svc_group: a list of strings for each service group
      db: network and service definitions

    Returns:
      results: a list of tuples for each service defined, in the format:
                       (service name, "<port>/<protocol>")
    """
    results = []
    for svc in svc_group:
        port = db.GetService(svc)
        results.append((svc, port))
    return results


def get_services(options, db):
    """Finds any services with that include a specific port/protocol pair.

    Args:
      options: the options sent to the script
      db: network and service definitions

    Returns:
      port, protocol, results as tuple in the format:
      (port, protocol, list of the services containing this pair)
    """
    results = []
    port, protocol = options.port
    # swap values if they were passed in wrong order
    if port.isalpha() and protocol.isdigit():
        port, protocol = protocol, port
    results = db.GetPortParents(port, protocol)
    return port, protocol, results


def EntryPoint():
    main(cli_options())


if __name__ == '__main__':
    EntryPoint()



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Command line interface to aclcheck library."""

import pathlib
from argparse import ArgumentParser, RawTextHelpFormatter

from aerleon.lib import aclcheck, naming, policy, yaml
from aerleon.utils import config


def main():
    _parser = ArgumentParser(
        prog='aclcheck',
        description=aclcheck.AclCheck.__doc__.splitlines()[0],
        formatter_class=RawTextHelpFormatter,
    )
    _parser.add_argument(
        '-p',
        '--policy-file',
        '--policy_file',
        dest='pol',
        help='The policy file to examine.',
        required=True,
    )
    _parser.add_argument(
        '--definitions-directory',
        '--definitions_directory',
        dest='definitions_directory',
        help='The directory where network and service definition files can be found.',
    )
    _parser.add_argument(
        '--base-directory',
        '--base_directory',
        dest='base_directory',
        help='The base directory to use when resolving policy include paths.',
    )
    _parser.add_argument(
        '--config-file',
        '--config_file',
        dest='config_file',
        help='Change the location searched for the configuration YAML file.',
    )
    _parser.add_argument('-d', '--destination', dest='destination_ip', help='Destination IP.')
    _parser.add_argument(
        '-s',
        '--source',
        dest='source_ip',
        help='Source IP.',
    )
    _parser.add_argument(
        '--proto',
        '--protocol',
        dest='protocol',
        help='Protocol (tcp, udp, icmp, etc.)',
    )
    _parser.add_argument(
        '--dport',
        '--destination-port',
        '--destination_port',
        dest='destination_port',
        help='Destination port.',
    )
    _parser.add_argument(
        '--sport', '--source-port', '--source_port', dest='source_port', help='Source port.'
    )
    FLAGS = _parser.parse_args()

    default_flags = {
        'base_directory': './policies',
        'definitions_directory': './def',
        'pol': None,
        'config_file': None,
        'destination_ip': '200.1.1.1',
        'source_ip': 'any',
        'protocol': 'any',
        'destination_port': '80',
        'source_port': '1025',
    }

    configs = {}
    configs.update(default_flags)

    # Pull common fields from aerleon.yml / --config_file
    # If all common fields are set on the command line, skip this step
    common_flags = frozenset(['base_directory', 'definitions_directory'])
    if not all(getattr(FLAGS, flag, None) for flag in common_flags):
        try:
            common_configs = config.load_config(config_file=FLAGS.config_file)
            common_configs = {
                key: value for key, value in common_configs.items() if key in common_flags
            }
            configs.update(common_configs)
        except config.ConfigFileError as e:
            exit(f"Error: {e}")

    configs.update({flag: value for flag, value in vars(FLAGS).items() if value})

    defs = naming.Naming(configs['definitions_directory'])

    with open(configs['pol']) as f:
        conf = f.read()

    if pathlib.Path(configs['pol']).suffix in ['.yaml', '.yml']:
        policy_obj = yaml.ParsePolicy(
            conf,
            base_dir=configs['base_directory'],
            filename=configs['pol'],
            definitions=defs,
        )
    else:
        policy_obj = policy.ParsePolicy(conf, defs, base_dir=configs['base_directory'])
    check = aclcheck.AclCheck(
        policy_obj,
        src=configs['source_ip'],
        dst=configs['destination_ip'],
        sport=configs['source_port'],
        dport=configs['destination_port'],
        proto=configs['protocol'],
    )
    print(str(check))


if __name__ == '__main__':
    main()



# Copyright 2014 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Discontinuous subnet mask summarizer."""
from __future__ import annotations

import collections

from aerleon.lib import nacaddr
from aerleon.lib.nacaddr import IPv4, IPv6


class DSMNet:
    """Hold IP address information for the purposes of DSM summarization.

    ipaddr maintainers explicitly declared that they will not
    support discontinuous subnet masks, hence this is required.
    """

    def __init__(self, address: int, netmask: int, text: str = '') -> None:
        """Creates DSMNet.

        Args:
          address: network address as int.
          netmask: subnet mask as int.
          text: text comment.
        """
        self.address = address
        self.netmask = netmask
        self.text = text

    def __hash__(self) -> int:
        return hash(str(self.address) + str(self.netmask))

    def __eq__(self, other: DSMNet) -> bool:
        try:
            return self.address == other.address and self.netmask == other.netmask
        except AttributeError:
            return NotImplemented

    def __ne__(self, other: DSMNet) -> bool:
        eq = self.__eq__(other)
        if eq is NotImplemented:
            return NotImplemented
        return not eq

    def __le__(self, other: DSMNet) -> bool:
        gt = self.__gt__(other)
        if gt is NotImplemented:
            return NotImplemented
        return not gt

    def __ge__(self, other: DSMNet) -> bool:
        lt = self.__lt__(other)
        if lt is NotImplemented:
            return NotImplemented
        return not lt

    def __lt__(self, other: DSMNet) -> bool:
        try:
            if self.address != other.address:
                return self.address < other.address
        except AttributeError:
            return NotImplemented
        return False

    def __gt__(self, other: DSMNet) -> bool:
        try:
            if self.address != other.address:
                return self.address > other.address
        except AttributeError:
            return NotImplemented
        return False

    def __str__(self) -> str:
        return f"{self.address} {self.netmask}"

    def MergeText(self, text: str = '') -> str:
        """Returns self.text joined with optional text.

        Don't join the text if it's already contained in self.text.

        Args:
          text: text to be combined with self.text.

        Returns:
          Combined text.
        """
        if self.text:
            if text and text not in self.text:
                return f"{self.text}, {text}"
            return self.text
        else:
            return text


def ToDottedQuad(net: DSMNet, negate: bool = False, nondsm: bool = False) -> tuple[str, str]:
    """Turns a DSMNet object into decimal dotted quad tuple.

    Args:
      net: DSMNet object.
      negate: if subnet mask should be negated (and become wildcard).
      nondsm: if mask should be generated in prefixlen when non-DSM.

    Returns:
      tuple (decimal dotted address, decimal dotted mask).

    Raises:
      ValueError: if address is larger than 32 bits or mask is not exactly
        0 or 32 bits.
    """
    if net.address.bit_length() > 32:
        raise ValueError('Addresses larger than 32 bits ' 'are currently not supported.')
    if net.netmask.bit_length() not in (0, 32):
        raise ValueError('Subnet masks other than 0 or 32 ' 'are currently not supported.')
    if negate:
        netmask = ~net.netmask
    else:
        netmask = net.netmask

    return (
        (_Int32ToDottedQuad(net.address), _PrefixlenForNonDSM(netmask))
        if nondsm
        else (_Int32ToDottedQuad(net.address), _Int32ToDottedQuad(netmask))
    )


def _PrefixlenForNonDSM(intmask: int) -> str:
    """Turns 32 bit integer into dotted decimal with JunOS friendly.

    Args:
      intmask: 32 bit integer.

    Returns:
      A string in dotted decimal or prefixlen format.
    """
    dotmask = _Int32ToDottedQuad(intmask)

    if dotmask == '255.255.255.255':
        return '32'

    bitmask = f'{intmask:032b}'

    prefixlen = 0
    while bitmask[prefixlen] == '1':
        prefixlen += 1

    return dotmask if int(bitmask[prefixlen:], 2) else str(prefixlen)


def _Int32ToDottedQuad(num: int) -> str:
    """Turns 32 bit integer into dotted decimal notation.

    Args:
      num: 32 bit integer.

    Returns:
      Integer as a string in dotted decimal notation.

    """
    octets = []
    for _ in range(4):
        octet = num & 0xFF
        octets.insert(0, str(octet))
        num >>= 8
    return '.'.join(octets)


def _NacaddrNetToDSMNet(net: IPv4 | IPv6) -> DSMNet:
    """Converts nacaddr.IPv4 or nacaddr.IPv6 object into DSMNet object.

    Args:
      net: nacaddr.IPv4 or nacaddr.IPv6 object.

    Returns:
      DSMNet object.
    """

    # left shift number of subnet mask bits, then leftshift until
    # full length of address reached
    address_as_int = int(net.network_address)
    netmask_as_int = ((1 << net.prefixlen) - 1) << (net.max_prefixlen - net.prefixlen)
    return DSMNet(address_as_int, netmask_as_int, net.text)


def _ToPrettyBinaryFormat(num: int) -> str:
    """Prettily formatted string of binary representation of suplied number.

    Useful for debugging.

    Args:
      num: number to be prettily formatted

    Returns:
      prettily formatted string
    """
    # like ipaddr make assumption that this is ipv4
    byte_strings = []
    while num > 0 or len(byte_strings) < 4:
        byte_strings.append(f'{num & 0xFF:08b}')
        num >>= 8
    return ' '.join(reversed(byte_strings))


def Summarize(nets: list[IPv4 | IPv6]) -> list[DSMNet]:
    """Summarizes networks while allowing for discontinuous subnet mask.

    Args:
      nets: list of nacaddr.IPv4 or nacaddr.IPv6 objects.
          Address family can be mixed, however there is no support for rendering
          anything other than IPv4.

    Returns:
      sorted list of DSMIPNet objects.
    """

    result = []
    optimized_nets = nacaddr.CollapseAddrList(nets)
    nets_by_netmask = collections.defaultdict(list)
    # group nets by subnet mask
    for net in optimized_nets:
        nets_by_netmask[net.prefixlen].append(_NacaddrNetToDSMNet(net))
    for nets in nets_by_netmask.values():
        result.extend(_SummarizeSameMask(nets))
    return sorted(result)


def _SummarizeSameMask(nets: list[DSMNet]) -> list[DSMNet]:
    """Summarizes networks while allowing for discontinuous subnet mask.

    Args:
      nets: list of unique, summarized DSMNet objects with the same netmask.

    Returns:
      sorted list of DSMNet objects that are discontinuously summarized.
    """

    # singletons can not possible be paired and are our result
    singletons = []
    # combinetons can potentially be paired
    combinetons = nets

    while combinetons:
        current_nets = combinetons
        combinetons = []
        while current_nets:
            current_net = current_nets.pop(0)
            # look for pair net, but keep index handy
            for pair_net_index, pair_net in enumerate(current_nets):
                xored_address = current_net.address ^ pair_net.address
                # For networks with the same network mask:
                # check if they have exactly one bit difference
                # or they are "a pair".
                if (
                    current_net.netmask == pair_net.netmask
                    and (xored_address & (xored_address - 1) == 0)
                    and xored_address > 0
                ):
                    # if pair was found, remove both, add paired up network
                    # to combinetons for next run and move along
                    # otherwise this network can never be paired
                    current_nets.pop(pair_net_index)
                    new_netmask = current_net.netmask ^ xored_address
                    # summarize supplied networks into one using discontinuous
                    # subnet mask.
                    combinetons.append(
                        DSMNet(
                            min(current_net.address, pair_net.address),
                            new_netmask,
                            current_net.MergeText(pair_net.text),
                        )
                    )
                    break
            else:
                singletons.append(current_net)
    return singletons



"""Builds Policy objects from a plain Python object representation."""

import enum
import sys
import typing
from dataclasses import dataclass, field
from typing import Annotated

from absl import logging

from aerleon.lib.policy import (
    FLEXIBLE_MATCH_RANGE_ATTRIBUTES,
    FLEXIBLE_MATCH_START_OPTIONS,
    Header,
    Policy,
    Target,
    Term,
    VarType,
)
from aerleon.lib.recognizers import (
    RecognizerContext,
    RecognizerKeywordResult,
    RecognizerValueResult,
    TComposition,
    TList,
    TListStr,
    TListStrCollapsible,
    TSection,
    TUnion,
    TValue,
)

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired, Required
else:
    from typing import NotRequired, Required

if sys.version_info < (3, 12):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict

if typing.TYPE_CHECKING:
    from datetime import datetime

    from aerleon.lib import naming


# NOTE: TypedDict is just a normal dictionary, it is not a class. This syntax tells the type checker
# what keys need to be present in the given dictionary.

WordList: TypeAlias = "str | list[str]"


class VPNValue(TypedDict):
    name: str
    policy: 'NotRequired[str]'


PolicyTerm = TypedDict(
    'PolicyTerm',
    {
        "comment": str,
        "name": "Required[str]",
        "action": WordList,
        "address": WordList,
        "address-exclude": WordList,
        "restrict-address-family": str,
        "counter": str,
        "expiration": "datetime.date | str",
        "destination-address": WordList,
        "destination-exclude": WordList,
        "destination-fqdn": WordList,
        "destination-port": WordList,
        "destination-prefix": WordList,
        "filter-term": str,
        "forwarding-class": "list[str]",
        "forwarding-class-except": "list[str]",
        "logging": WordList,
        "log-limit": str,
        "log-name": str,
        "loss-priority": str,
        "option": "list[str]",
        "owner": str,
        "policer": str,
        "port": WordList,
        "precedence": "str | list[int | str]",
        "protocol": "str | list[int | str]",
        "protocol-except": "str | list[int | str]",
        "qos": str,
        "pan-application": WordList,
        "routing-instance": str,
        "source-address": WordList,
        "source-exclude": WordList,
        "source-fqdn": WordList,
        "source-port": WordList,
        "source-prefix": WordList,
        "ttl": "int | str",
        "verbatim": "dict[str, str]",
        "packet-length": "int | str",
        "fragment-offset": "int | str",
        "hop-limit": "int | str",
        "icmp-type": WordList,
        "icmp-code": "str | list[int | str]",
        "ether-type": WordList,
        "traffic-class-count": str,
        "traffic-type": WordList,
        "dscp-set": "int | str",
        "dscp-match": "int | str | list[int | str]",
        "dscp-except": "int | str | list[int | str]",
        "next-ip": str,
        "flexible-match-range": "dict[str, int | str]",
        "source-prefix-except": WordList,
        "destination-prefix-except": WordList,
        "encapsulate": str,
        "port-mirror": str,
        "destination-zone": WordList,
        "source-zone": WordList,
        "vpn": VPNValue,
        "source-tag": WordList,
        "destination-tag": WordList,
        "priority": "int | str",
        "source-interface": str,
        "destination-interface": str,
        "platform": WordList,
        "platform-exclude": WordList,
        "target-resources": "list[str | list[str]]",
        "target-service-accounts": WordList,
        "timeout": "int | str",
    },
    total=False,
)


class PolicyInclude(TypedDict):
    include: str


PolicyFilterHeader = TypedDict(
    'PolicyFilterHeader',
    {
        "targets": "Required[dict[str, str]]",
        "comment": str,
        "apply-groups": WordList,
        "apply-groups-except": WordList,
    },
    total=False,
)


TermsList = list[PolicyTerm | PolicyInclude]


class PolicyFilter(TypedDict):
    header: PolicyFilterHeader
    terms: TermsList


class PolicyDict(TypedDict):
    filters: 'list[PolicyFilter]'
    filename: NotRequired[str]


RawTarget = Annotated[
    str,
    """
    RawTarget contains a partially evaluated representation of platform-specific target options.
    This type may change to accomodate complex target-specific options.
    """,
]


@dataclass
class RawFilterHeader:
    """Contains a partially evaluated representation of a filter header.

    Attributes:
        targets: A mapping from target name to target-specific options.
        kvs: A mapping containing any other key/value pairs in the header.
    """

    targets: dict[str, RawTarget]
    kvs: dict[str, typing.Any] = field(default_factory=dict)


@dataclass
class RawTerm:
    """Contains a partially evaluated representation of a filter term.

    Attributes:
        name: The term name.
        kvs A mapping containing all other key/value pairs in the term.
    """

    name: str
    kvs: dict[str, typing.Any] = field(default_factory=dict)


@dataclass
class RawFilter:
    """Contains a partially evaluated representation of a policy filter.

    Attributes:
        header: The filter header.
        terms: The filter term list.
    """

    header: RawFilterHeader
    terms: "list[RawTerm]"


@dataclass
class RawPolicy:
    """Contains a partially evaluated representation of a Policy.

    Attributes:
        filename: filename for this policy.
        filters: list of filters in this policy.
    """

    filename: str
    filters: "list[RawFilter]"


class PolicyBuilder:
    """PolicyBuilder produces a Policy object from a RawPolicy. This allows a Policy
    object to be constructed without having or creating a .pol file.

    PolicyBuilder works by constructing Term, Header, Target, and Policy models
    directly. These models expect to be constructed incrementally by ply during
    ply's parsing procedure using methods like Term.AddObject(), Header.AddObject()
    and Policy.AddFilter(). PolicyBuilder walks through the given RawPolicy
    and incrementally constructs the models using the same API used by ply.

    PolicyBuilder is responsible for recognizing the keys and values in the
    given RawPolicy. Unexpected keywords or unexpected value formats are ignored
    with a warning. The recognizer process also parses information from values.

    PolicyBuilder is responsible for normalizing values across equivalent inputs.
    A .pol file is essentially composed of strings, but RawPolicy is a
    structure of Python native objects, where date values might be represented
    as a Python date and numeric values might be represented as numbers.

    PolicyBuilder also carries configuration forward to the Policy model. The
    Policy model may optionally perform optimization or shade checking during
    model construction. This behavior is configured through module-global
    variables within policy.py. policy.FromBuilder() will read this configuration
    off the PolicyBuilder and pass it through the module-global channel.

    Usage:

    policy_builder = PolicyBuilder(raw_policy, definitions, optimize, shade_check)
    return policy.FromBuilder(policy_builder)

    Attributes:
        raw_policy: The given RawPolicy, an intermediate representation
            containing filters, headers, terms, etc.
        definitions: The collection of network and service name
            definitions used to resolve names found in the RawPolicy.
        optimize: Enables policy optimization if true.
        shade_check: Enable shade checking if true.
    """

    raw_policy: RawPolicy
    definitions: "naming.Naming"
    optimize: bool
    shade_check: bool

    def __init__(
        self,
        policy_dict: PolicyDict,
        definitions: "naming.Naming",
        optimize=False,
        shade_check=False,
    ):
        def PolicyDictToRawPolicy(input_policy):
            raw_filters = []
            input_filters = input_policy["filters"]
            filename = input_policy.get("filename")
            for filter in input_filters:
                filter_header = filter["header"]
                header_targets = filter_header["targets"]
                raw_filter_header = RawFilterHeader(targets=header_targets, kvs=filter_header)

                raw_terms = []
                filter_terms = filter["terms"]
                for term in filter_terms:
                    raw_term = RawTerm(name=term["name"], kvs=term)
                    raw_terms.append(raw_term)

                raw_filters.append(RawFilter(header=raw_filter_header, terms=raw_terms))

            return RawPolicy(filename=filename, filters=raw_filters)

        self.raw_policy = PolicyDictToRawPolicy(policy_dict)
        self.definitions = definitions
        self.optimize = optimize
        self.shade_check = shade_check

    def BuildPolicy(self):
        """Build a Policy model from a RawPolicy.

        Do not call this method directly. See "Usage" in the class docstring.
        """
        policy_model = None
        obj_calls = [self._BuildFilter(policy_filter) for policy_filter in self.raw_policy.filters]
        for obj in obj_calls:
            if policy_model is None:
                policy_model = Policy(*obj)
            else:
                policy_model.AddFilter(*obj)
        policy_model.filename = self.raw_policy.filename
        return policy_model

    def _BuildFilter(self, policy_filter: RawFilter):
        kvs_parsed = {}
        header_kvs = policy_filter.header.kvs
        header_kvs["targets"] = policy_filter.header.targets

        # Run the builtin recognizers against each key/value pair.
        for keyword, value in header_kvs.items():
            recognizer_context = RecognizerContext(
                policy=None,
                filter=policy_filter,
                header=policy_filter.header,
                target=None,
                term=None,
                keyword=keyword,
                value=value,
            )
            kw_recognizer_result = HeaderBuiltinRecognizer.RecognizeKeyword(recognizer_context)
            if not kw_recognizer_result.recognized:
                logging.warning(f"Unexpected term keyword {keyword}")
                continue
            val_recognizer_result = HeaderBuiltinRecognizer.RecognizeKeywordValue(
                recognizer_context
            )
            if not val_recognizer_result.recognized:
                logging.warning(f"Unrecognized value format for {keyword}: {value}")
                continue
            for repr_key, repr_value in val_recognizer_result.valueKV.items():
                kvs_parsed[repr_key] = repr_value

        # Construct the Header instance and its Target.
        header = Header()
        for target, target_options in kvs_parsed["targets"].items():
            target_args = [target]
            target_args.extend(target_options)
            target = Target(target_args)
            header.AddObject(target)

        # Add all remaining fields to the header.
        for keyword, value in kvs_parsed.items():
            if keyword == 'targets':
                continue

            # Build the call sequence for Header.AddObject for each field.
            # The correct call sequence depends on the keyword.
            # See class docstring for _CallType for details.
            obj_calls = _Builtin.FromKeyword(keyword).AddObjectCallSequence(value)
            for obj in obj_calls:
                header.AddObject(obj)

        # Now construct all Term instances and asseble the call sequence for Policy / Policy.AddObject
        return (header, [self._BuildTerm(term) for term in policy_filter.terms])

    def _BuildTerm(self, term: RawTerm):
        # It is an error for a term to be empty or to only contain a name.
        if not len(term.kvs):
            raise TypeError("Term must have at least one keyword.")

        if term.name is None:
            raise TypeError("Term must have a name.")

        # Validate name using the builtin recognizers.
        term_name_recognizer_result = TermBuiltinRecognizer.RecognizeKeywordValue(
            RecognizerContext(
                policy=None,
                filter=None,
                header=None,
                target=None,
                term=term,
                keyword='name',
                value=term.name,
            )
        )
        if not term_name_recognizer_result.recognized:
            raise TypeError("Invalid term name.")

        # Run recognizers over all term items. This validates, parses and normalizes each value.
        kvs_parsed = {}
        for keyword, value in term.kvs.items():
            recognizer_context = RecognizerContext(
                policy=None,
                filter=None,
                header=None,
                target=None,
                term=term,
                keyword=keyword,
                value=value,
            )
            kw_recognizer_result = TermBuiltinRecognizer.RecognizeKeyword(recognizer_context)
            if not kw_recognizer_result.recognized:
                logging.warning(f"Unexpected term keyword {keyword}")
                continue
            val_recognizer_result = TermBuiltinRecognizer.RecognizeKeywordValue(recognizer_context)
            if not val_recognizer_result.recognized:
                logging.warning(f"Unrecognized value format for {keyword}: {value}")
                continue
            for repr_key, repr_value in val_recognizer_result.valueKV.items():
                kvs_parsed[repr_key] = repr_value

        term.kvs = kvs_parsed

        # Construct a Term model. This is not especially straightforward.
        #
        # At this point the values in term.kvs have been parsed and normalized.
        #
        # policy.Term() will crash if initialized without a VarType object.
        # This will initialize the Term on the first iteration cycle over
        # the kvs and call AddObject on subsequent passes.
        #
        # More details are provided in the class docstring for _CallType.

        term_model = None  # Will initialize on first cycle

        for keyword, value in term.kvs.items():
            if keyword == 'name':
                continue

            obj_calls = _Builtin.FromKeyword(keyword).AddObjectCallSequence(value)

            for obj in obj_calls:
                if term_model is None:
                    term_model = Term(obj)
                else:
                    term_model.AddObject(obj)
        term_model.name = term.name
        return term_model


# ### BUILTINS ###
# The following section deals with the recognition and normalization of built-in keywords.

BUILTIN_SPEC: "dict[str, TValue | TComposition]" = {
    # fmt: off
    'apply-groups':               TListStrCollapsible,
    'apply-groups-except':        TListStrCollapsible,
    'comment':                    TValue.AnyString,
    'targets':                    TSection(of=[(TValue.WordString, TListStrCollapsible)]),
    'name':                       TValue.WordString,
    'action':                     TListStrCollapsible,
    'address':                    TListStrCollapsible,
    'address-exclude':            TListStrCollapsible,
    'restrict-address-family':    TValue.WordString,
    'counter':                    TValue.WordString,
    'expiration':                 TValue.YearMonthDay,
    'destination-address':        TListStrCollapsible,
    'destination-exclude':        TListStrCollapsible,
    'destination-fqdn':       TListStrCollapsible,
    'destination-port':           TListStrCollapsible,
    'destination-prefix':         TListStrCollapsible,
    'filter-term':                TValue.WordString,
    'forwarding-class':           TList(of=TValue.WordString),
    'forwarding-class-except':    TList(of=TValue.WordString),
    'logging':                    TListStrCollapsible,
    'log-limit':                  TValue.LogLimit,
    'log-name':                   TValue.AnyString,
    'loss-priority':              TValue.WordString,
    'option':                     TListStr,
    'owner':                      TValue.WordString,
    'policer':                    TValue.WordString,
    'port':                       TListStrCollapsible,
    'precedence':                 TList(of=TValue.Integer, collapsible=True),
    'protocol':                   TList(of=TUnion(of=[TValue.Integer, TValue.WordString]), collapsible=True),
    'protocol-except':            TList(of=TUnion(of=[TValue.Integer, TValue.WordString]), collapsible=True),
    'qos':                        TValue.WordString,
    'pan-application':            TListStrCollapsible,
    'routing-instance':           TValue.WordString,
    'source-address':             TListStrCollapsible,
    'source-exclude':             TListStrCollapsible,
    'source-fqdn':            TListStrCollapsible,
    'source-port':                TListStrCollapsible,
    'source-prefix':              TListStrCollapsible,
    'ttl':                        TValue.Integer,
    'verbatim':                   TSection(of=[(TValue.WordString, TValue.AnyString)]),
    # juniper specific.           
    'packet-length':              TUnion(of=[TValue.Integer, TValue.IntegerRange]),
    'fragment-offset':            TUnion(of=[TValue.Integer, TValue.IntegerRange]),
    'hop-limit':                  TUnion(of=[TValue.Integer, TValue.IntegerRange]),
    'icmp-type':                  TListStrCollapsible,
    'icmp-code':                  TList(of=TValue.Integer, collapsible=True),
    'ether-type':                 TListStrCollapsible,
    'traffic-class-count':        TValue.WordString,
    'traffic-type':               TListStrCollapsible,
    'dscp-set':                   TValue.DSCP,
    'dscp-match':                 TList(of=TUnion(of=[TValue.DSCP, TValue.DSCPRange]), collapsible=True),
    'dscp-except':                TList(of=TUnion(of=[TValue.DSCP, TValue.DSCPRange]), collapsible=True),
    'next-ip':                    TValue.WordString,
    'flexible-match-range':       TSection(of=[(TValue.WordString, TUnion(of=[TValue.Hex, TValue.Integer, TValue.WordString]))]),
    'source-prefix-except':       TListStrCollapsible,
    'destination-prefix-except':  TListStrCollapsible,
    'encapsulate':                TValue.WordString,
    'port-mirror':                TValue.WordString,
    # srx specific                
    'destination-zone':           TListStrCollapsible,
    'source-zone':                TListStrCollapsible,
    'vpn':                        TSection(of=[('name', TValue.WordString), ('policy', TValue.WordString)]),
    # gce specific                
    'source-tag':                 TListStrCollapsible,
    'destination-tag':            TListStrCollapsible,
    'priority':                   TValue.Integer,
    # iptables specific           
    'source-interface':           TValue.WordString,
    'destination-interface':      TValue.WordString,
    'platform':                   TListStrCollapsible,
    'platform-exclude':           TListStrCollapsible,
    'target-resources':           TList(of=TValue.TargetResourceTuple),
    'target-service-accounts':    TListStrCollapsible,
    'timeout':                    TValue.Integer,
    # fmt: on
}


class _CallType(enum.Enum):
    """_CallType enumerates the calling conventions used by Term.AddObject()
    and Header.AddObject() in policy.py.

    Term.AddObject() is an unusual and complex function that adapts the ply
    parser to the Policy data model. It is the only method of attaching data
    to a Term model instance. Specifically it mirrors the numerous grammar rule
    function implementations (roughly one grammar rule is defined for each
    VarType type).

    Term.AddObject() accepts a single parameter, obj, containing either a single
    VarType object or a list of VarType objects. The semantics of how values
    are represented within 'obj' vary by the type of the VarType object. These
    semantics broadly fall into three cases, enumerated by this class.

    Attributes:
        SingleValue: AddObject should only be called zero or one times for this
            VarType type. 'obj' should contain a single VarType instance.
        SingleList: AddObject should only be called zero or one times for this
            VarType type. 'obj' should contain a list of VarType instances.
        MultiCall: AddObject should be called once per value. 'obj' should contain
            a single VarType instance.

    N.b. A few VarType types expect the VarType object value to be a list
    in order to represent tuple-like values. It does not affect the AddObject
    calling convention.
    """

    SingleValue = enum.auto()
    SingleList = enum.auto()
    MultiCall = enum.auto()


class _Builtin:
    """_Builtin represents all possible built-in key-value pairs found
    in Policy Term and Header sections.

    This class sort of serves to reverse-engineer Header.AddObject and
    Term.AddObject. See the _CallType class docstring for more details on
    what we are working around.
    """

    # fmt: off
    BUILTINS = {
        # KEYNAME:                    (CALL_CONVENTION,        VAR_TYPE)
        'name':                       (None,                   None),
        'targets':                    (None,                   None),
        'next-ip':                    (_CallType.SingleValue,  VarType.NEXT_IP),
        'restrict-address-family':    (_CallType.SingleValue,  VarType.RESTRICT_ADDRESS_FAMILY),
        'owner':                      (_CallType.SingleValue,  VarType.OWNER),
        'expiration':                 (_CallType.SingleValue,  VarType.EXPIRATION),
        'loss-priority':              (_CallType.SingleValue,  VarType.LOSS_PRIORITY),
        'routing-instance':           (_CallType.SingleValue,  VarType.ROUTING_INSTANCE),
        'precedence':                 (_CallType.SingleValue,  VarType.PRECEDENCE),
        'counter':                    (_CallType.SingleValue,  VarType.COUNTER),
        'encapsulate':                (_CallType.SingleValue,  VarType.ENCAPSULATE),
        'port-mirror':                (_CallType.SingleValue,  VarType.PORT_MIRROR),
        'traffic-class-count':        (_CallType.SingleValue,  VarType.TRAFFIC_CLASS_COUNT),
        'icmp-type':                  (_CallType.SingleValue,  VarType.ICMP_TYPE),
        'icmp-code':                  (_CallType.SingleValue,  VarType.ICMP_CODE),
        'log-name':                   (_CallType.SingleValue,  VarType.LOG_NAME),
        'log-limit':                  (_CallType.SingleValue,  VarType.LOG_LIMIT),
        'policer':                    (_CallType.SingleValue,  VarType.POLICER),
        'priority':                   (_CallType.SingleValue,  VarType.PRIORITY),
        'qos':                        (_CallType.SingleValue,  VarType.QOS),
        'packet-length':              (_CallType.SingleValue,  VarType.PACKET_LEN),
        'fragment-offset':            (_CallType.SingleValue,  VarType.FRAGMENT_OFFSET),
        'hop-limit':                  (_CallType.SingleValue,  VarType.HOP_LIMIT),
        'source-interface':           (_CallType.SingleValue,  VarType.SINTERFACE),
        'destination-interface':      (_CallType.SingleValue,  VarType.DINTERFACE),
        'timeout':                    (_CallType.SingleValue,  VarType.TIMEOUT),
        'dscp-set':                   (_CallType.SingleValue,  VarType.DSCP_SET),
        'ttl':                        (_CallType.SingleValue,  VarType.TTL),
        'filter-term':                (_CallType.SingleValue,  VarType.FILTER_TERM),
        'vpn':                        (_CallType.SingleValue,  VarType.VPN),
        'source-address':             (_CallType.SingleList,   VarType.SADDRESS),
        'destination-address':        (_CallType.SingleList,   VarType.DADDRESS),
        'destination-fqdn':       (_CallType.SingleList,   VarType.DESTINATION_FQDN),
        'source-fqdn':            (_CallType.SingleList,   VarType.SOURCE_FQDN),
        'address':                    (_CallType.SingleList,   VarType.ADDRESS),
        'source-exclude':             (_CallType.SingleList,   VarType.SADDREXCLUDE),
        'destination-exclude':        (_CallType.SingleList,   VarType.DADDREXCLUDE),
        'address-exclude':            (_CallType.SingleList,   VarType.ADDREXCLUDE),
        'port':                       (_CallType.SingleList,   VarType.PORT),
        'source-port':                (_CallType.SingleList,   VarType.SPORT),
        'destination-port':           (_CallType.SingleList,   VarType.DPORT),
        'protocol':                   (_CallType.SingleList,   VarType.PROTOCOL),
        'protocol-except':            (_CallType.SingleList,   VarType.PROTOCOL_EXCEPT),
        'option':                     (_CallType.SingleList,   VarType.OPTION),
        'source-prefix':              (_CallType.SingleList,   VarType.SPFX),
        'source-prefix-except':       (_CallType.SingleList,   VarType.ESPFX),
        'destination-prefix':         (_CallType.SingleList,   VarType.DPFX),
        'destination-prefix-except':  (_CallType.SingleList,   VarType.EDPFX),
        'ether-type':                 (_CallType.SingleList,   VarType.ETHER_TYPE),
        'traffic-type':               (_CallType.SingleList,   VarType.TRAFFIC_TYPE),
        'forwarding-class':           (_CallType.SingleList,   VarType.FORWARDING_CLASS),
        'forwarding-class-except':    (_CallType.SingleList,   VarType.FORWARDING_CLASS_EXCEPT),
        'pan-application':            (_CallType.SingleList,   VarType.PAN_APPLICATION),
        'platform':                   (_CallType.SingleList,   VarType.PLATFORM),
        'platform-exclude':           (_CallType.SingleList,   VarType.PLATFORMEXCLUDE),
        'source-tag':                 (_CallType.SingleList,   VarType.STAG),
        'destination-tag':            (_CallType.SingleList,   VarType.DTAG),
        'target-resources':           (_CallType.SingleList,   VarType.TARGET_RESOURCES),
        'target-service-accounts':    (_CallType.SingleList,   VarType.TARGET_SERVICE_ACCOUNTS),
        'source-zone':                (_CallType.SingleList,   VarType.SZONE),
        'destination-zone':           (_CallType.SingleList,   VarType.DZONE),
        'apply-groups':               (_CallType.SingleList,   VarType.APPLY_GROUPS),
        'apply-groups-except':        (_CallType.SingleList,   VarType.APPLY_GROUPS_EXCEPT),
        'dscp-match':                 (_CallType.SingleList,   VarType.DSCP_MATCH),
        'dscp-except':                (_CallType.SingleList,   VarType.DSCP_EXCEPT),
        'flexible-match-range':       (_CallType.SingleList,   VarType.FLEXIBLE_MATCH_RANGE),
        'action':                     (_CallType.MultiCall,    VarType.ACTION),
        'comment':                    (_CallType.MultiCall,    VarType.COMMENT),
        'logging':                    (_CallType.MultiCall,    VarType.LOGGING),
        'verbatim':                   (_CallType.MultiCall,    VarType.VERBATIM),
    }
    # fmt: on

    def __init__(self, keyname: str, call_convention: _CallType, var_type: VarType):
        self.keyname = keyname
        self.call_convention = call_convention
        self.var_type = var_type

    @classmethod
    def FromKeyword(cls, keyname: str):
        """Construct a Builtin instance from keyname.

        Args:
            keyname (string): Builtin keyname.

        Returns:
            A Builtin instance.

        Raises:
            KeyError if no builtin data is registered for argument keyname.
        """
        return cls(keyname, *cls.BUILTINS[keyname])

    @property
    def recognizer(self) -> "TValue | TComposition":
        """The recognizer specific to this Builtin instance."""
        return BUILTIN_SPEC[self.keyname]

    def AddObjectCallSequence(self, value: typing.Any):
        """Construct a calling sequence for Term.AddObject or Header.AddObject for
        this builtin type.

        Args:
            value (Any): Variable value.

        Returns:
            A list where each item contains the args list for a single call
            to AddObject.
        """
        if self.call_convention == _CallType.SingleValue:
            return [VarType(self.var_type, value)]  # One call
        elif self.call_convention == _CallType.SingleList:
            return [[VarType(self.var_type, item) for item in value]]  # One call
        elif self.call_convention == _CallType.MultiCall:
            return [VarType(self.var_type, item) for item in value]  # One call per item


class BuiltinRecognizer:
    """BuiltinRecognizer recognizes and normalizes built-in keywords and values.

    Attributes:
        ALLOWED_BUILTIN_KEYS: Subclasses of BuiltinRecognizer can set this
            class attribute to control the behavior of the default implementation
            of recognizeKeyword().
    """

    ALLOWED_BUILTIN_KEYS = ()

    @classmethod
    def RecognizeKeyword(cls, context: RecognizerContext) -> RecognizerKeywordResult:
        """Examines the given RecognizerContext and determines whether
        the keyword (context.keyword) is valid in context.

        The default implementation checks whether the keyword is found in ALLOWED_BUILTIN_KEYS.

        Args:
            context: The given RecognizerContext. This contains the keyword in question
                and additional context about where the keyword is located in the policy.

        Returns:
            A RecognizerKeywordResult with the 'recognized' field set to True if and only if
            the keyword is valid in the current context. The 'securityCritical' field is always
            set to False and should be ignored.
        """
        securityCritical = False
        recognized = context.keyword in cls.ALLOWED_BUILTIN_KEYS
        return RecognizerKeywordResult(recognized=recognized, securityCritical=securityCritical)

    @classmethod
    def RecognizeKeywordValue(cls, context: RecognizerContext) -> RecognizerValueResult:
        """Examines the given RecognizerContext and determines whether
        the key/value pair is valid in context.

        Args:
            context: The given RecognizerContext. This contains the keyword in question,
                the value in question and additional context about where the keyword
                is located in the policy.

        Returns:
            A RecognizerKeywordResult with the 'recognized' field set to True if and only if
            the key/value pair is valid in the current context. The 'securityCritical' field is always
            set to False and should be ignored. If the key/value pair is valid the valueKV field will
            contain a dict with a single key (the input keyword) mapped to a parsed and normalized
            representation of the input value.

            Value parsing is performed by executing a small program specific to the input keyword
            against the input value. Each program is written using the TValue / TComposite
            classes defined in recognizers.py . For example, for the keyword 'packet-length', the input value
            must be either an integer, a string containing an integer, or a string containing two
            integers separated by a dash (a range) and the recognizer program would be
            "TUnion(of=[TValue.Integer, TValue.IntegerRange])".

            Value parsing does provide some normalization across values, for example TList understands that
            a list might be given as a Python list or a string containing space-separated values and will
            produce the same parsed output for both cases. Lists of integers, dates, DSCP values, and others
            all have a little flexibility in representation that is normalized by TValue / TComposite.

            Subclasses can implement the normalizeValues class method to apply further adjustments to the output.
        """

        try:
            recognizer = _Builtin.FromKeyword(context.keyword).recognizer
            repr = recognizer.Recognize(context.value)
        except KeyError:
            return RecognizerValueResult(recognized=False)
        except TypeError:
            return RecognizerValueResult(recognized=False)

        # Last-minute validations and normalization
        try:
            if context.keyword == "comment" and isinstance(repr, str):
                # Plain term objects should specify a comment as a single string,
                # potentially a multi-line string. The Policy model will represent
                # the comment as a list of lines, so we perform the transformation here.
                repr = repr.splitlines()
            elif context.keyword == 'verbatim':
                # Verbatim tables must be presented as a list of 2-tuples.
                repr = repr.items()
            elif recognizer == TValue.Integer:
                # All builtins that accept a single integer must pass that integer as a string
                # into the model.
                repr = str(repr)
            elif recognizer == TUnion(of=[TValue.Integer, TValue.IntegerRange]):
                # All builtins that accept an integer/integer range must pass a string
                # into the model.
                repr = str(repr)
            elif context.keyword in ("protocol", "protocol-except"):
                # Builtins that accept a list of words or integers must represent any integers
                # as strings in the mode.
                repr = [str(value) for value in repr]

            # Allow for additional section-specific normalizations
            repr = cls._NormalizeValues(context, repr)
        except TypeError:
            return RecognizerValueResult(recognized=False)

        # We have recognized the keyword, asserted the representation matches the tokenizer
        # and assembed a native object representation.
        return RecognizerValueResult(recognized=True, valueKV={context.keyword: repr})

    @classmethod
    def _NormalizeValues(cls, _context: RecognizerContext, repr: typing.Any) -> typing.Any:
        """Subclasses of BuiltinRecognizer can implement this method to adjust the output
        of recognizeKeywordValue.

        Args:
            context: The RecognizerContext given to recognizeKeywordValue.
            repr: The value representation produced by recognizeKeywordValue.

        Returns:
            An adjusted version of the input value "repr". The default implementation returns
            the input value unchanged.

        Raises:
            TypeError: The input value does not match requirements. The default implementation
            will never raise this error.
        """

        return repr


class HeaderBuiltinRecognizer(BuiltinRecognizer):
    """Implements BuiltinRecognizer for filter header values. See BuiltinRecognizer."""

    ALLOWED_BUILTIN_KEYS = frozenset(
        {
            'apply-groups',
            'apply-groups-except',
            'comment',
            'targets',
        }
    )


class TermBuiltinRecognizer(BuiltinRecognizer):
    """Implements BuiltinRecognizer for filter term values. See BuiltinRecognizer."""

    ALLOWED_BUILTIN_KEYS = frozenset(
        {
            'name',
            'action',
            'address',
            'address-exclude',
            'restrict-address-family',
            'comment',
            'counter',
            'expiration',
            'destination-address',
            'destination-exclude',
            'destination-port',
            'destination-prefix',
            'filter-term',
            'forwarding-class',
            'forwarding-class-except',
            'logging',
            'log-limit',
            'log-name',
            'loss-priority',
            'option',
            'owner',
            'policer',
            'port',
            'precedence',
            'protocol',
            'protocol-except',
            'qos',
            'pan-application',
            'routing-instance',
            'source-address',
            'source-exclude',
            'source-port',
            'source-prefix',
            'ttl',
            'verbatim',
            'packet-length',
            'fragment-offset',
            'hop-limit',
            'icmp-type',
            'icmp-code',
            'ether-type',
            'traffic-class-count',
            'traffic-type',
            'dscp-set',
            'dscp-match',
            'dscp-except',
            'next-ip',
            'flexible-match-range',
            'source-prefix-except',
            'destination-prefix-except',
            'encapsulate',
            'port-mirror',
            'destination-zone',
            'source-zone',
            'vpn',
            'source-tag',
            'destination-tag',
            'priority',
            'source-interface',
            'destination-interface',
            'platform',
            'platform-exclude',
            'target-resources',
            'target-service-accounts',
            'timeout',
            'destination-fqdn',
            'source-fqdn',
        }
    )

    @classmethod
    def _NormalizeValues(cls, context: RecognizerContext, repr):
        """See BuiltinRecognizer.normalizeValues()."""

        if context.keyword == "flexible-match-range":
            # Flexible match range values are validated during the lex/yacc parsing phase.
            # TODO(jb) This validation step can be performed once within the Term model in the
            # SanityCheck step and removed from both the lex/yacc code and this builder code.
            new_repr = {}
            for key, value in repr.items():
                if key not in FLEXIBLE_MATCH_RANGE_ATTRIBUTES:
                    raise TypeError(f"Flexible match range: {key} is not a valid attribute")
                if key == "match-start":
                    if value not in FLEXIBLE_MATCH_START_OPTIONS:
                        raise TypeError(f"Flexible match range: {key} value is not valid")
                # per Juniper, max bit length is 32
                elif key == "bit-length":
                    if int(value) not in list(range(33)):
                        raise TypeError(f"Flexible match range: {key} value is not valid")
                # per Juniper, max bit offset is 7
                elif key == "bit-offset":
                    if int(value) not in list(range(8)):
                        raise TypeError(f"Flexible match range: {key} value is not valid")
                # per Juniper, offset can be up to 256 bytes
                elif key == "byte-offset":
                    if int(value) not in list(range(256)):
                        raise TypeError(f"Flexible match range: {key} value is not valid")
                # Policy model expects all values as strings
                new_repr[key] = str(value)
            # Policy model expects each kv represented as a length-2 tuple
            repr = new_repr.items()

        elif context.keyword == "vpn":
            if 'name' not in repr:
                raise TypeError("VPN: keyword 'name' is mising.")
            # Policy model expects each kv represented as a length-2 tuple.
            # The policy name field should be given as an empty string if not provided
            # by the user.
            repr = (repr['name'], repr.get('policy', ''))
        return repr



# Copyright 2016 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Windows advfirewall policy generator."""

import string
from typing import Union

from aerleon.lib import windows
from aerleon.lib.nacaddr import IPv4, IPv6


class Term(windows.Term):
    """Generate windows advfirewall policy terms."""

    _PLATFORM = 'windows_advfirewall'
    CMD_PREFIX = 'netsh advfirewall firewall '

    # Logging:
    # netsh advfirewall>set allprofiles logging allowedconnections enable"
    # netsh advfirewall>set allprofiles logging droppedconnections enable"

    # 'in' or 'out'
    _DIR_ATOM = string.Template('dir=${dir}')
    # 'local' or 'remote'
    _ADDR_ATOM = string.Template('${dir}ip=${addr}')
    _PORT_ATOM = string.Template('${dir}port=${port}')
    # any | Integer | icmpv4 | icmpv6 | icmpv4:type,code | icmpv6:type,code
    # | tcp | udp
    _PROTO_ATOM = string.Template('protocol=${protocol}')
    # 'allow' or 'block'
    _ACTION_ATOM = string.Template('action=${action}')

    _RULE_FORMAT = string.Template(
        'add rule name=${name} enable=yes ' 'interfacetype=any ${atoms}'
    )

    _ACTION_TABLE = {
        'accept': 'allow',
        'deny': 'block',
        'reject': 'block',
    }

    def _HandleIcmpTypes(
        self, icmp_types: list[str], protocols: list[str]
    ) -> tuple[list[str], list[str]]:
        # advfirewall actually puts this in the protocol spec, eg.:
        # icmpv4 | icmpv6 | icmpv4:type,code | icmpv6:type,code
        types = ['']
        if icmp_types:
            types = self.NormalizeIcmpTypes(self.term.icmp_type, protocols, self.af)
            # NormalizeIcmpTypes enforces this the af/ip version match:
            icmp_prefix = 'icmpv4'
            if self.af == 'inet6':
                icmp_prefix = 'icmpv6'

            if types:
                protocols = []
                for typ in types:
                    protocols.append(f'{icmp_prefix}:{typ},any')
                types = ['']

        # fixup for icmp v4
        for i in range(len(protocols)):
            if protocols[i] == 'icmp':
                protocols[i] = 'icmpv4'

        return (types, protocols)

    def _HandlePorts(
        self, src_ports: list[tuple[int, int]], dst_ports: list[tuple[int, int]]
    ) -> tuple[list[str], list[str]]:
        return ([self._ComposePortString(src_ports)], [self._ComposePortString(dst_ports)])

    def _CartesianProduct(
        self,
        src_addr: list[Union[IPv4, IPv6]],
        dst_addr: list[Union[IPv4, IPv6]],
        protocol: list[str],
        unused_icmp_types: list[str],
        src_port: list[str],
        dst_port: list[str],
        ret_str: list[str],
    ) -> None:
        # At least advfirewall supports port ranges, unlike windows ipsec,
        # so the src and dst port lists will always be one element long.
        for saddr in src_addr:
            for daddr in dst_addr:
                for proto in protocol:
                    ret_str.append(
                        self._ComposeRule(
                            saddr, daddr, proto, src_port[0], dst_port[0], self.term.action[0]
                        )
                    )

    def _ComposeRule(
        self, srcaddr: IPv4, dstaddr: IPv4, proto: str, srcport: str, dstport: str, action: str
    ) -> str:
        """Convert the given parameters into a netsh add rule string."""
        atoms = []
        src_label = 'local'
        dst_label = 'remote'

        # We assume a default direction of OUT, but if it's IN, the Windows
        # advfirewall changes around the remote and local labels.
        if 'in' == self.filter.lower():
            src_label = 'remote'
            dst_label = 'local'

        atoms.append(self._DIR_ATOM.substitute(dir=self.filter))

        if srcaddr.prefixlen == 0:
            atoms.append(self._ADDR_ATOM.substitute(dir=src_label, addr='any'))
        else:
            atoms.append(self._ADDR_ATOM.substitute(dir=src_label, addr=str(srcaddr)))

        if dstaddr.prefixlen == 0:
            atoms.append(self._ADDR_ATOM.substitute(dir=dst_label, addr='any'))
        else:
            atoms.append(self._ADDR_ATOM.substitute(dir=dst_label, addr=str(dstaddr)))

        if srcport:
            atoms.append(self._PORT_ATOM.substitute(dir=src_label, port=srcport))
        if dstport:
            atoms.append(self._PORT_ATOM.substitute(dir=dst_label, port=dstport))

        if proto:
            if proto == 'vrrp':
                proto = '112'
            elif proto == 'ah':
                proto = '51'
            elif proto == 'hopopt':
                proto = '0'
            atoms.append(self._PROTO_ATOM.substitute(protocol=proto))

        atoms.append(self._ACTION_ATOM.substitute(action=self._ACTION_TABLE[action]))

        return self.CMD_PREFIX + self._RULE_FORMAT.substitute(
            name=self.term_name, atoms=' '.join(atoms)
        )

    def _ComposePortString(self, ports: list[tuple[int, int]]) -> str:
        """Convert the list of ports tuples into a multiport range string."""
        if not ports:
            return ''

        multiports = []
        for start, end in ports:
            if start == end:
                multiports.append(str(start))
            else:
                multiports.append(f"{start!s}-{end!s}")
        return ','.join(multiports)


class WindowsAdvFirewall(windows.WindowsGenerator):
    """Generates filters and terms from provided policy object."""

    _PLATFORM = 'windows_advfirewall'
    _TERM = Term



# Copyright 2015 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple, line-oriented parser for Caprica policies.

This parser differs from the default parser in that it preserves the original
structure and defers resolving referents to the user. This is useful for
analyzing policy structures and their use of naming data. It happens to discard
inline comments but preservers line-level comments. Fields expected to have
"naming" values are stored as a set without order or line breaks retained.
"""

from __future__ import annotations

from absl import logging


class Field:
    """A name-value assignment within a block."""

    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self) -> str:
        t = type(self)
        f = 'UNKNOWN'
        for k, v in field_map.items():
            if t == v:
                f = k
                break
        indent = len(f) + 5
        return '{}::{}'.format(f, self.ValueStr().replace('\n', '\n' + ' ' * indent))

    def __eq__(self, o: Target) -> bool:
        if not isinstance(o, self.__class__):
            return False
        return self.value == o.value

    def __ne__(self, o: Target) -> bool:
        return not self == o

    def Append(self, value: str) -> None:
        self.value += value

    def ValueStr(self) -> str:
        return self.value


class IntegerField(Field):
    def __init__(self, value: str) -> None:
        super().__init__(value)
        try:
            _ = int(value)
        except ValueError:
            raise ValueError(f'Invalid integer field: "{self!s}"')


class NamingField(Field):
    """A naming field is one that refers to names in used in naming.py."""

    def __init__(self, value: str) -> None:
        super().__init__(value)
        self.value = self.ParseString(value)

    def ParseString(self, value: str) -> set[str]:
        """Split and validate a string value into individual names."""
        parts = set(value.split())
        for p in parts:
            self.ValidatePart(p)
        return parts

    def ValidatePart(self, part: str) -> None:
        """Validate that a string smells like a naming.py name."""
        for c in part:
            if c not in '-_.' and not c.isdigit() and not c.isupper():
                raise ValueError(f'Invalid name reference: "{part}"')

    def Append(self, value: str) -> None:
        """Split, validate, and add name contained within a string."""
        parts = self.ParseString(value)
        self.value.update(parts)

    def ValueStr(self) -> str:
        """Return the value as a series of lines no longer than 60 chars each."""
        values = sorted(self.value)
        line_wrap = 60
        length = 0
        line_buf = []
        value_buf = []
        for v in values:
            if length + len(v) > line_wrap:
                value_buf.append(' '.join(line_buf))
                length = 0
                line_buf = []
            else:
                if line_buf:
                    length += 1
                line_buf.append(v)
                length += len(v)
        if line_buf:
            value_buf.append(' '.join(line_buf))
        return ' ' + '\n'.join(value_buf)


class Action(Field):
    """An action field."""


class Address(NamingField):
    """An address field."""


class Port(NamingField):
    """A port field."""


class Comment(Field):
    """A comment field."""

    def ValueStr(self):
        # Comments should align with the string contents, after the leading
        # quotation mark.
        return self.value.replace('\n', '\n ')


class Counter(Field):
    """A counter field."""


class Encapsulate(Field):
    """An encapsulate field."""


class DestinationAddress(Address):
    """A destination-address field."""


class DestinationExclude(Address):
    """A destination-exclude field."""


class DestinationInterface(Field):
    """A destination-interface field."""


class DestinationPort(Port):
    """A destination-port field."""


class DestinationPrefix(Field):
    """A destination-prefix field."""


class DestinationPrefixExcept(Field):
    """A destination-prefix-except field."""


class DestinationTag(Field):
    """A destination tag field."""


class DestinationZone(Field):
    """A destination-zone field."""


class DscpMatch(Field):
    """A dscp-match field."""


class DscpSet(Field):
    """A dscp-set field."""


class EtherType(Field):
    """An ether-type field."""


class Expiration(Field):
    """An expiration field."""


class FragmentOffset(Field):
    """A fragment-offset field."""


class ForwardingClass(Field):
    """A forwarding-class field."""


class ForwardingClassExcept(Field):
    """A forwarding-class-except field."""


class IcmpCode(Field):
    """A icmp-code field."""


class IcmpType(Field):
    """A icmp-type field."""


class Logging(Field):
    """A logging field."""


class LossPriority(Field):
    """A loss-priority field."""


class Option(Field):
    """An Option field."""


class Owner(Field):
    """An owner field."""


class NextIP(Field):
    """An owner field."""


class PacketLength(Field):
    """A packet-length field."""


class Platform(Field):
    """A platform field."""


class PlatformExclude(Field):
    """A platform-exclude field."""


class Policer(Field):
    """A rate-limit-icmp field."""


class PortMirror(Field):
    """A port-mirror field."""


class Precedence(Field):
    """A precedence field."""


class Protocol(Field):
    """A Protocol field."""


class ProtocolExcept(Field):
    """A protocol-except field."""


class Qos(Field):
    """A rate-limit-icmp field."""


class PANApplication(Field):
    """A rate-limit-icmp field."""


class RoutingInstance(Field):
    """A routing-instance field."""


class SourceAddress(Address):
    """A source-address field."""


class SourceExclude(Address):
    """A source-exclude field."""


class SourceInterface(Field):
    """A source-interface field."""


class SourcePort(Port):
    """A source-port field."""


class SourcePrefix(Field):
    """A source-prefix field."""


class SourcePrefixExcept(Field):
    """A source-prefix-except field."""


class SourceTag(Field):
    """A source tag field."""


class SourceZone(Field):
    """A source-zone field."""


class Target(Field):
    """A target field."""


class Timeout(IntegerField):
    """A timeout field."""


class TrafficType(Field):
    """A traffic-type field."""


class TrafficClassCount(Field):
    """A traffic-class-count field."""


class Verbatim(Field):
    """A verbatim field."""


class Vpn(Field):
    """A vpn field."""


destination_address_fields = (DestinationAddress, DestinationExclude, DestinationPrefix)

field_map = {
    'action': Action,
    'address': Address,
    'comment': Comment,
    'counter': Counter,
    'destination-address': DestinationAddress,
    'destination-exclude': DestinationExclude,
    'destination-interface': DestinationInterface,
    'destination-port': DestinationPort,
    'destination-prefix': DestinationPrefix,
    'destination-prefix-except': DestinationPrefixExcept,
    'destination-tag': DestinationTag,
    'destination-zone': DestinationZone,
    'dscp-match': DscpMatch,
    'dscp-set': DscpSet,
    'ether-type': EtherType,
    'expiration': Expiration,
    'fragment-offset': FragmentOffset,
    'forwarding-class': ForwardingClass,
    'forwarding-class-except': ForwardingClassExcept,
    'icmp-code': IcmpCode,
    'icmp-type': IcmpType,
    'logging': Logging,
    'loss-priority': LossPriority,
    'option': Option,
    'owner': Owner,
    'next-ip': NextIP,
    'packet-length': PacketLength,
    'platform': Platform,
    'platform-exclude': PlatformExclude,
    'policer': Policer,
    'port': Port,
    'port-mirror': PortMirror,
    'precedence': Precedence,
    'protocol': Protocol,
    'protocol-except': ProtocolExcept,
    'qos': Qos,
    'pan-application': PANApplication,
    'routing-instance': RoutingInstance,
    'source-address': SourceAddress,
    'source-exclude': SourceExclude,
    'source-interface': SourceInterface,
    'source-port': SourcePort,
    'source-prefix': SourcePrefix,
    'source-prefix-except': SourcePrefixExcept,
    'source-tag': SourceTag,
    'source-zone': SourceZone,
    'target': Target,
    'timeout': Timeout,
    'traffic-class-count': TrafficClassCount,
    'traffic-type': TrafficType,
    'verbatim': Verbatim,
    'vpn': Vpn,
    'encapsulate': Encapsulate,
}


class Block:
    """A section containing fields."""

    def __init__(self) -> None:
        self.fields = []

    def __iter__(self):
        return iter(self.fields)

    def __getitem__(self, i):
        return self.fields[i]

    def __str__(self) -> str:
        buf = []
        buf.append(type(self).__name__.lower())
        buf.append(' ')
        if self.Name():
            buf.append(self.Name())
            buf.append(' ')
        buf.append('{')  # }
        buf.append('\n')
        for field in self.fields:
            buf.append('  ')
            buf.append(str(field))
            buf.append('\n')
        buf.append('}')
        buf.append('\n')
        return ''.join(buf)

    def AddField(self, field) -> None:
        if not issubclass(type(field), Field):
            raise TypeError(f'{field} not subclass of Field.')
        self.fields.append(field)

    def FieldsWithType(self, f_type: type[Comment]) -> list[Comment]:
        if not issubclass(f_type, Field):
            raise TypeError(f'{f_type} not subclass of Field.')
        return [x for x in self.fields if isinstance(x, f_type)]

    def Match(self, match_fn):
        """Yield the fields and their indices for which match_fn is True."""
        for i, f in enumerate(self.fields):
            if match_fn(f):
                yield i, f

    def Name(self) -> str:
        return ''

    def __eq__(self, o: Header | Term) -> bool:
        if not isinstance(o, self.__class__):
            return False
        if len(self.fields) != len(o.fields):
            return False
        for mine, theirs in zip(self.fields, o.fields):
            logging.debug('testing "%s" vs "%s"', mine, theirs)
            if mine != theirs:
                return False
        return True

    def __ne__(self, o):
        return not self == o


class Header(Block):
    """A header block."""


class Term(Block):
    """A policy term."""

    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name

    def Name(self) -> str:
        return self.name

    def __eq__(self, o: Term) -> bool:
        if not super().__eq__(o):
            return False
        return self.name == o.name

    def Describe(self):
        """Return a human-readable description of the term."""
        verbatims = self.FieldsWithType(Verbatim)
        if verbatims:
            return f'Verbatim: {verbatims}'

        handled = set()
        handled.update(self.FieldsWithType(Comment))

        pieces = []
        actions = self.FieldsWithType(Action)
        if len(actions) != 1:
            raise ValueError('No action or multiple actions.')
        handled.update(actions)
        pieces.append(actions[0].value.title() + ' traffic')

        protocols = self.FieldsWithType(Protocol)
        all_protocols = set()
        if protocols:
            handled.update(protocols)
            for protocol in protocols:
                all_protocols.update(protocol.value.split())
            pieces.append('using ' + ' or '.join(sorted(all_protocols)))

        icmp_code = self.FieldsWithType(IcmpCode)
        all_icmp_code = set()
        if icmp_code:
            handled.update(icmp_code)
            for code in icmp_code:
                all_icmp_code.update(code.value.split())
            pieces.append(f"(ICMP code {', '.join(sorted(all_icmp_code))})")

        icmp_types = self.FieldsWithType(IcmpType)
        all_icmp_types = set()
        if icmp_types:
            handled.update(icmp_types)
            for icmp_type in icmp_types:
                all_icmp_types.update(icmp_type.value.split())
            pieces.append(f"(ICMP types {', '.join(sorted(all_icmp_types))})")

        sources = self.FieldsWithType(SourceAddress)
        if sources:
            handled.update(sources)
            pieces.append('originating from')
            all_sources = set()
            for source in sources:
                all_sources.update(source.value)
            pieces.append(', '.join(sorted(all_sources)))

        source_ports = self.FieldsWithType(SourcePort)
        if source_ports:
            handled.update(source_ports)
            if sources:
                pieces.append('using port')
            else:
                pieces.append('originating port')
            all_sources = set()
            for source in source_ports:
                all_sources.update(source.value)
            pieces.append(', '.join(sorted(all_sources)))

        destinations = self.FieldsWithType(DestinationAddress)
        if destinations:
            handled.update(destinations)
            pieces.append('destined for')
            all_destinations = set()
            for destination in destinations:
                all_destinations.update(destination.value)
            pieces.append(', '.join(sorted(all_destinations)))

        destination_ports = self.FieldsWithType(DestinationPort)
        if destination_ports:
            handled.update(destination_ports)
            if destinations:
                pieces.append('on port')
            else:
                pieces.append('destined for port')
            all_destinations = set()
            for destination in destination_ports:
                all_destinations.update(destination.value)
            pieces.append(', '.join(sorted(all_destinations)))

        vpns = self.FieldsWithType(Vpn)
        if vpns:
            handled.update(vpns)
            pieces.append('via VPNs')
            pieces.append(','.join(x.value for x in vpns))

        # Ignore some fields
        for ignored_type in (Expiration, Owner):
            ignored_fields = self.FieldsWithType(ignored_type)
            if ignored_fields:
                handled.update(ignored_fields)

        for field in self:
            if field not in handled:
                raise ValueError('Uncovered field: ' + str(field))
        return ' '.join(pieces)


class BlankLine:
    """A blank line."""

    def __str__(self):
        return '\n'

    def __eq__(self, o: BlankLine) -> bool:
        return isinstance(o, self.__class__)

    def __ne__(self, o):
        return not self == o


class CommentLine:
    """A comment in the file."""

    def __init__(self, data: str) -> None:
        self.data = data

    def __str__(self):
        return str(self.data) + '\n'

    def __eq__(self, o: CommentLine) -> bool:
        if not isinstance(o, self.__class__):
            return False
        return self.data == o.data

    def __ne__(self, o):
        return not self == o


class Include:
    """A reference to another policy definition."""

    def __init__(self, identifier: str) -> None:
        self.identifier = identifier

    def __str__(self):
        return f'#include {self.identifier}'

    def __eq__(self, o: Include) -> bool:
        if not isinstance(o, self.__class__):
            return False
        return self.identifier == o.identifier

    def __ne__(self, o):
        return not self == o


class Policy:
    """An ordered list of headers, terms, comments, blank lines and includes."""

    def __init__(self, identifier: str | None) -> None:
        self.identifier = identifier
        self.members = []

    def AddMember(self, member) -> None:
        m_type = type(member)
        if m_type not in (Include, CommentLine, BlankLine) and not issubclass(m_type, Block):
            raise TypeError(f'{m_type} must be a Block, CommentLine, BlankLine, or Include')
        self.members.append(member)

    def __str__(self) -> str:
        return ''.join(str(x) for x in self.members)

    def __iter__(self):
        return iter(self.members)

    def __getitem__(self, i):
        return self.members[i]

    def Match(self, match_fn):
        """Yield the members and their indices for which match_fn is True."""
        for i, m in enumerate(self.members):
            if match_fn(m):
                yield i, m

    def MatchFields(self, block_match_fn, field_match_fn):
        for match_idx, m in self.Match(block_match_fn):
            if not isinstance(m, Block):
                continue
            for field_idx, f in m.Match(field_match_fn):
                yield match_idx, field_idx, f


class PolicyParser:
    """Parse a policy object from a data buffer."""

    def __init__(self, data: str, identifier: str) -> None:
        self.data = data
        self.identifier = identifier
        self.block_in_progress = None
        self.policy = None

    def Parse(self) -> Policy:
        """Do the needful."""
        self.policy = Policy(self.identifier)
        for line in self.data.split('\n'):
            line = line.strip()
            logging.debug('Processing line: "%s"', line)
            if self.block_in_progress:
                self.ParseInBlock(line)
            else:
                self.ParseTopLevel(line)
        if self.block_in_progress:
            raise ValueError(f'Unexpected EOF reading "{self.block_in_progress}"')
        return self.policy

    def ParseTopLevel(self, line: str) -> None:
        """Parse a line not nested within a block."""
        if line == '':  # pylint: disable=g-explicit-bool-comparison
            self.policy.AddMember(BlankLine())
            return
        if line.startswith('#'):
            if line.startswith('#include '):
                self.ParseIncludeLine(line)
                return
            self.ParseCommentLine(line)
            return
        if line.startswith('header {') or line.startswith('header{'):  # }
            self.ParseHeaderLine(line)
            return
        if line.startswith('term '):
            self.ParseTermLine(line)
            return
        raise ValueError(f'Unhandled top-level line {line}')

    def ParseCommentLine(self, line: str) -> None:
        """Parse a line with a line level comment."""
        if self.block_in_progress:
            raise ValueError(f'Found comment line in block: {line}')
        self.policy.AddMember(CommentLine(line))

    def ParseIncludeLine(self, line: str) -> None:
        """Parse an #include line refering to another file."""
        if self.block_in_progress:
            raise ValueError(f'Found include line in block: {line}')
        line_parts = line.split()
        if len(line_parts) < 2:
            raise ValueError(f'Invalid include: {line}')
        inc_ref = line_parts[1]
        if '#' in inc_ref:
            inc_ref, _ = inc_ref.split('#', 1)
        self.policy.AddMember(Include(inc_ref))

    def ParseHeaderLine(self, line: str) -> None:
        """Parse a line beginning a header block."""
        if self.block_in_progress:
            raise ValueError(f'Nested blocks not allowed: {line}')
        self.block_in_progress = Header()

    def ParseTermLine(self, line: str) -> None:
        """Parse a line beginning a term block."""
        if self.block_in_progress:
            raise ValueError(f'Nested blocks not allowed: {line}')
        line_parts = line.split()

        # Some terms don't have a space after the name
        if '{' in line_parts[1]:  # }
            brace_idx = line_parts[1].index('{')  # }
            line_parts[1] = line_parts[1][:brace_idx]
        else:
            if not line_parts[2].startswith('{'):  # }
                raise ValueError(f'Invalid term line: {line}')
        term_name = line_parts[1]
        self.block_in_progress = Term(term_name)

    def ParseInBlock(self, line: str) -> None:
        """Parse a line when inside a block definition."""
        if line == '' or line.startswith('#'):  # pylint: disable=g-explicit-bool-comparison
            return
        if '::' in line:
            self.ParseField(line)
            return
        if line.startswith('}'):
            self.policy.AddMember(self.block_in_progress)
            self.block_in_progress = None
            return
        if self.block_in_progress is not None:
            self.block_in_progress.fields[-1].Append('\n' + line)

    def ParseField(self, line: str) -> None:
        """Parse a line containing a block field."""
        name, value = line.split('::', 1)
        name = name.strip().lower()
        f_type = field_map.get(name)
        if not f_type:
            raise ValueError(f'Invalid field line: {line}')
        self.block_in_progress.AddField(f_type(value))



# Copyright 2023 The Capirca Project Authors All Rights Reserved.
# Copyright 2023 VMware, Inc. SPDX-License-Identifier: Apache-2.0
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""nsxt generator."""

import json

from absl import logging

from aerleon.lib import aclgenerator

_ACTION_TABLE = {
    'accept': 'ALLOW',
    'deny': 'DROP',
    'reject': 'REJECT',
    'reject-with-tcp-rst': 'REJECT',  # tcp rst not supported
}

_NSXT_SUPPORTED_KEYWORDS = [
    'name',
    'action',
    'resource_type',
    'display_name',
    'marked_for_delete',
    'overridden',
    'rule_id',
    'sequence_number',
    'sources_excluded',
    'destinations_excluded',
    'source_groups',
    'destination_groups',
    'services',
    'profiles',
    'logged',
    'scope',
    'disabled',
    'notes',
    'direction',
    'tag',
    'ip_protocol',
    'is_default',
    'protocol',
    'destination_port',
    'source_address',
    'destination_address',
    'source_port',
]

_PROTOCOLS = {1: 'ICMPv4', 6: 'TCP', 17: 'UDP', 58: 'ICMPv6'}


# generic error class
class Error(Exception):
    """Generic error class."""

    pass


class UnsupportedNsxtAccessListError(Error):
    """Raised when we're give a non named access list."""

    pass


class NsxtAclTermError(Error):
    """Raised when there is a problem in a nsxt access list."""

    pass


class NsxtDuplicateTermError(Error):
    """Raised when there is a duplicate."""

    pass


class NsxtUnsupportedCriteriaOperatorError(Error):
    """Raised when an unsupported criteria comparison operator is encountered."""

    pass


class NsxtUnsupportedManyPoliciesError(Error):
    """Raised when there are many policies/headers specified."""

    pass


class ServiceEntries:
    """Represents service entries for a rule."""

    def __init__(self, protocol, source_ports, destination_ports, icmp_types):
        """Setting things up.

        Args:
          protocol: str, protocol.
          source_ports: str list or none, the source port.
          destination_ports: str list or none, the destination port.
          icmp_types: icmp-type numeric specification (if any).
        """
        self.protocol = protocol
        self.source_ports = source_ports
        self.destination_ports = destination_ports
        self.icmp_types = icmp_types

    def get(self):
        """Returns list of services."""
        # Handle ICMP and ICMPv6
        if self.protocol == 1 or self.protocol == 58:
            service = {
                'protocol': _PROTOCOLS[self.protocol],
                'resource_type': 'ICMPTypeServiceEntry',
            }
            if not self.icmp_types:
                return [service]

            # Handle ICMP types
            services = []
            for icmp_type in self.icmp_types:
                new_service = service.copy()
                new_service['icmp_type'] = icmp_type
                services.append(new_service)
                return services

        # Handle TCP and UDP
        elif self.protocol == 6 or self.protocol == 17:
            service = {
                'l4_protocol': _PROTOCOLS[self.protocol],
                'resource_type': 'L4PortSetServiceEntry',
            }

            # Handle Layer 4 Ports
            if self.source_ports:
                source_ports = [f'{p[0]}-{p[1]}' for p in self.source_ports]
                service['source_ports'] = source_ports

            if self.destination_ports:
                destination_ports = [f'{p[0]}-{p[1]}' for p in self.destination_ports]
                service['destination_ports'] = destination_ports
            return [service]
        else:
            return []


class Term(aclgenerator.Term):
    """Creates a single ACL Term for NSX-T."""

    def __init__(self, term, filter_type, applied_to=None, af=4):
        self.term = term
        # Our caller should have already verified the address family.
        assert af in (4, 6)
        self.af = af
        self.filter_type = filter_type
        self.applied_to = applied_to

    def __str__(self):
        """Convert term to a rule string.

        Returns:
          A rule as a string.

        """
        # Verify platform specific terms. Skip whole term if platform does not
        # match.
        if self.term.platform:
            if 'nsxt' not in self.term.platform:
                return ''
        if self.term.platform_exclude:
            if 'nsxt' in self.term.platform_exclude:
                return ''

        term_keywords = self.term.__dict__
        unsupported_keywords = []
        for key in term_keywords:
            if term_keywords[key]:
                if ('translated' not in key) and (key not in _NSXT_SUPPORTED_KEYWORDS):
                    unsupported_keywords.append(key)
        if unsupported_keywords:
            logging.warning(
                'WARNING: The keywords %s in Term %s are not supported ' 'in Nsxt ',
                unsupported_keywords,
                self.term.name,
            )

        name = self.term.name

        notes = ''
        if self.term.comment:
            notes = '\n'.join(self.term.comment)

        action = 'ALLOW'
        if self.term.action:
            action = _ACTION_TABLE.get(self.term.action[0])

        source_address = ['ANY']
        if self.term.source_address:
            source_address = []
            for i in self.term.source_address:
                source_address.append(str(i))

        destination_address = ['ANY']
        if self.term.destination_address:
            destination_address = []
            for i in self.term.destination_address:
                destination_address.append(str(i))

        rule = {
            'action': action,
            'resource_type': 'Rule',
            'display_name': name,
            'source_groups': source_address,
            'destination_groups': destination_address,
            # Set mandatory services to ANY, as service_entries will be used
            'services': ['ANY'],
            'profiles': ['ANY'],
            'scope': ['ANY'],
            'logged': bool(self.term.logging),
            'notes': notes,
            'direction': 'IN_OUT',
            'ip_protocol': 'IPV4_IPV6',
        }

        if self.term.protocol:
            icmp_types = []
            services = []
            if self.term.icmp_type:
                icmp_types = self.NormalizeIcmpTypes(
                    self.term.icmp_type, self.term.protocol, self.af
                )

            protocol = [self.PROTO_MAP.get(p) for p in self.term.protocol]
            for proto in protocol:
                if proto != 'any':
                    service = ServiceEntries(
                        proto, self.term.source_port, self.term.destination_port, icmp_types
                    )
                    services.extend(service.get())
            rule['service_entries'] = services

        return json.dumps(rule)


class Nsxt(aclgenerator.ACLGenerator):
    """nsxt rendering class.

      This class takes a policy object and renders the output into a syntax
      which is understood by nsxt policy.

    Attributes:
      pol: policy.Policy object

    Raises:
    UnsupportednsxtAccessListError: Raised when we're give a non named access
    list.

    """

    _PLATFORM = 'nsxt'
    _DEFAULT_PROTOCOL = 'ip'
    SUFFIX = '.nsxt'

    _OPTIONAL_SUPPORTED_KEYWORDS = {'expiration', 'logging'}
    _FILTER_OPTIONS_DICT = {}

    def _TranslatePolicy(self, pol, exp_info):
        self.nsxt_policies = []

        for header, terms in pol.filters:
            if self._PLATFORM not in header.platforms:
                continue

            filter_options = header.FilterOptions(self._PLATFORM)
            if len(filter_options) >= 2:
                filter_name = filter_options[1]

            self._ParseFilterOptions(filter_options)

            filter_type = ''
            applied_to = ''

            term_names = set()
            new_terms = []
            for term in terms:
                # Check for duplicate terms
                if term.name in term_names:
                    raise NsxtDuplicateTermError(f'There are multiple terms named: {term.name}')
                term_names.add(term.name)

                term.name = self.FixTermLength(term.name)

                new_terms.append(Term(term, filter_type, applied_to, 4))

            self.nsxt_policies.append((header, filter_name, [filter_type], new_terms))

    def _ParseFilterOptions(self, filter_options):
        """Parses the target in header for filter type, section_id and applied_to.

        Args:
          filter_options: list of remaining target options

        Returns:
          A dictionary that contains fields necessary to create the firewall
          rule.

        Raises:
          UnsupportedNsxtAccessListError: Raised when we're give a non named access
          list.
        """
        # check for filter type
        if not 2 <= len(filter_options) <= 5:
            raise UnsupportedNsxtAccessListError(
                'Invalid Number of options specified: %d. Required options '
                'are: filter type and section name. Platform: %s'
                % (len(filter_options), self._PLATFORM)
            )
        # mandatory section_name
        section_name = filter_options[0]
        # mandatory
        filter_type = filter_options[1]

        # a mixed filter outputs both ipv4 and ipv6 acls in the same output file
        good_filters = ['inet', 'inet6', 'mixed']

        # check if filter type is renderable
        if filter_type not in good_filters:
            raise UnsupportedNsxtAccessListError(
                'Access list type %s not supported by %s (good types: %s)'
                % (filter_type, self._PLATFORM, str(good_filters))
            )

        section_id = None
        applied_to = 'ANY'
        filter_opt_len = len(filter_options)

        if filter_opt_len > 2:
            for index in range(2, filter_opt_len):
                if index == 2 and filter_options[2] != 'securitygroup':
                    section_id = filter_options[2]
                    continue
                if filter_options[index] == 'securitygroup':
                    if index + 1 <= filter_opt_len - 1:
                        securitygroup = filter_options[index + 1]
                        if securitygroup[0] != '/':
                            securitygroup = f'/infra/domains/default/groups/{securitygroup}'
                        applied_to = securitygroup
                        break
                    else:
                        raise UnsupportedNsxtAccessListError(
                            f'Security Group Id is not provided for {self._PLATFORM}'
                        )

        self._FILTER_OPTIONS_DICT['section_name'] = section_name
        self._FILTER_OPTIONS_DICT['filter_type'] = filter_type
        self._FILTER_OPTIONS_DICT['section_id'] = section_id
        self._FILTER_OPTIONS_DICT['applied_to'] = applied_to

    def __str__(self):
        """Render the output of the nsxt policy."""
        if len(self.nsxt_policies) > 1:
            raise NsxtUnsupportedManyPoliciesError('Only one policy can be rendered')
        for _, _, _, terms in self.nsxt_policies:
            rules = [json.loads(str(term)) for term in terms]

        section_name = self._FILTER_OPTIONS_DICT['section_name']
        section_id = self._FILTER_OPTIONS_DICT['section_id']
        applied_to = self._FILTER_OPTIONS_DICT['applied_to']

        policy = {
            'rules': rules,
            'resource_type': 'SecurityPolicy',
            'display_name': section_name,
            'id': section_id if section_id is not None else section_name,
            'category': 'Application',
            'is_default': 'false',
            'scope': [applied_to],
        }

        return json.dumps(policy, indent=2)
